"""
Functions to creates random "true" region and site models for demodata.
"""
import geojson
import kwarray
import kwimage
import mgrs
import ubelt as ub
# from datetime import datetime as datetime_cls
from shapely.ops import unary_union
from kwarray.distributions import TruncNormal
from kwgis.demo.metrics_demo import demo_utils


class RegionModelGenerator:
    """
    Note:
        A good refactor to this random data design would be to have a class
        which holds all the parameters to generate random site / region models.
        For now we are using this as a placeholder to store constants.

    TODO:
        Transfer the functionality of `random_region_model` here.
    """
    ...


class SiteModelGenerator:
    """
    TODO:
        Transfer the functionality of random_site_model here.
    """
    SITE_PHASES = [
        "No Activity",
        "Site Preparation",
        "Active Construction",
        "Post Construction",
    ]


def random_region_model(region_id=None, region_poly=None, num_sites=3,
                        num_observations=5, p_observe=0.5, p_transition=0.15,
                        start_time=None, end_time=None,
                        with_renderables=True, site_poly=None, rng=None,
                        start_date=None, end_date=None, start_site_index=0):
    """
    Generate a random region model with random sites and observation support.

    The region model is generated simply by sampling a random polygon in
    geospace that isn't too big or too small.

    Then observation support simulates images that we have from this region.
    These supporting images might be from different simulated sensors or in
    different resolutions. They may contain partial coverage of the region.

    The sites are randomly sampled polygons inside the region and are assigned
    to specific observation as they would be if they were generated for real
    imagery. Sites have a simple model that allows them to evolve over time,
    change shape / size and their phase label.

    Args:
        region_id (str | None): Name of the region.
            If unspecified, a random one is created.

        region_poly (kwimage.Polygon | shapely.geometry.Polygon | None):
            if specified use this crs84 region polygon, otherwise make a random
            one.

        num_sites (int): number of random sites

        num_observations (int): number of random observations

        p_observe (float):
            the probability that a site model is annotated in any particular
            observation.

        p_transition (float):
            truth phase transition model. Currently just the probability
            the phase changes on any particular observation.

        with_renderables (bool): if False dont generate renderables.

        start_time (Any): min time coercable

        end_time (Any): max time coercable

        site_poly (kwimage.Polygon | shapely.geometry.Polygon | None):
            if specified, this polygon is used as the geometry for new site
            models. Note: all site models will get this geometry, so
            typically this is only used when num_sites=1.

        rng (int | str | RandomState | None) : seed or random number generator

        start_site_index (int):
            the index of the first generated site. Defaults to 0

    Returns:
        Tuple[geojson.FeatureCollection, List[geojson.FeatureCollection], List | None]:
            A region model and its corresponding site models and renderables

    Example:
        >>> from kwgis.demo.metrics_demo.demo_truth import *  # NOQA
        >>> region, sites, renderables = random_region_model(num_sites=2, num_observations=5, p_observe=0.5, rng=0)
        >>> print('region = {}'.format(ub.urepr(region, nl=4, precision=6, sort=0)))
        ...

        region = {
            'type': 'FeatureCollection',
            'features': [
                {
                    'type': 'Feature',
                    'geometry': {
                        'type': 'Polygon',
                        'coordinates': ...
                    },
                    'properties': {
                        'type': 'region',
                        'region_id': 'DR_R684',
                        'version': '2.4.3',
                        'mgrs': '51PXM',
                        'start_date': '2011-05-28',
                        'end_date': '2018-09-13',
                        'originator': 'demo-truth',
                        'model_content': 'annotation',
                        'comments': 'demo-data',
                    },
                },
                {
                    'type': 'Feature',
                    'geometry': {
                        'type': 'Polygon',
                        'coordinates': ...
                    },
                    'properties': {
                        'type': 'site_summary',
                        'status': 'positive_annotated',
                        'version': '2.0.1',
                        'site_id': 'DR_R684_0000',
                        'mgrs': '51PXM',
                        'start_date': '2015-03-16',
                        'end_date': '2018-09-13',
                        'score': 1,
                        'originator': 'demo',
                        'model_content': 'annotation',
                        'validated': 'True',
                        'misc_info': {'color': [0.551139, 1.000000, 0.000000]},
                    },
                },
                {
                    'type': 'Feature',
                    'geometry': {
                        'type': 'Polygon',
                        'coordinates': ...
                    },
                    'properties': {
                        'type': 'site_summary',
                        'status': 'positive_annotated',
                        'version': '2.0.1',
                        'site_id': 'DR_R684_0001',
                        'mgrs': '51PXM',
                        'start_date': '2011-05-28',
                        'end_date': '2018-09-13',
                        'score': 1,
                        'originator': 'demo',
                        'model_content': 'annotation',
                        'validated': 'True',
                        'misc_info': {'color': [1.000000, 0.367780, 0.000000]},
                    },
                },
            ],
        }
    """
    from kwutil import util_time
    rng = kwarray.ensure_rng(rng)

    if start_date is not None:
        raise ValueError('Use start_time instead of start_date')

    if end_date is not None:
        raise ValueError('Use end_time instead of end_date')

    if num_observations <= 0:
        raise ValueError('must have at least one observation')

    if region_id is None:
        region_id = 'DR_R{:03d}'.format(rng.randint(0, 1000))

    if region_poly is None:
        region_poly = demo_utils.random_geo_polygon(max_rt_area=10_000, rng=rng)

    region_poly = kwimage.Polygon.coerce(region_poly)
    region_geom = region_poly.to_shapely()

    lon, lat = region_poly.centroid
    mgrs_code = mgrs.MGRS().toMGRS(lat, lon, MGRSPrecision=0)

    # Determine how we made each observation
    observables = random_observables(num_observations, start_time=start_time,
                                     end_time=end_time, rng=rng)
    wld_polygon = region_poly
    for observation in observables:
        # Could be more complex here, but other code would have to change
        # Making assumptions everything is aligned in initial pass
        observation['mgrs_code'] = mgrs_code
        observation['wld_polygon'] = wld_polygon

    if end_time is None:
        end_time = observables[-1]['datetime']
    else:
        end_time = util_time.coerce_datetime(end_time)

    if start_time is None:
        start_time = observables[0]['datetime']
    else:
        start_time = util_time.coerce_datetime(start_time)

    # Define the region feature
    region_feature = geojson.Feature(
        properties={
            "type": "region",
            "region_id": region_id,
            "version": "2.4.3",
            "mgrs": mgrs_code,
            "start_date": start_time.date().isoformat(),
            "end_date": end_time.date().isoformat(),
            "originator": "demo-truth",
            "model_content": "annotation",
            "comments": "demo-data",
        },
        geometry=region_geom,
    )

    region_box = region_poly.bounding_box()
    region_corners = region_box.corners()

    # Create random site models within this region
    sites = []
    site_summaries = []
    for site_num in range(start_site_index, start_site_index + num_sites):
        site_id = f"{region_id}_{site_num:04d}"
        sitesum, site = random_site_model(
            region_id, site_id, region_corners, observables,
            site_poly=site_poly, p_observe=p_observe,
            p_transition=p_transition, rng=rng)
        site_summaries.append(sitesum)
        sites.append(site)

    region = geojson.FeatureCollection([region_feature] + site_summaries)

    # Create information about how we should render dummy images
    if not with_renderables:
        renderables = None
    else:
        renderables = []
        for frame_idx, observable in enumerate(observables):

            # For each observed date, generate information about the toy images we
            # "observed" and what visible data is in them.
            image_box = kwimage.Boxes([[0, 0, 800, 600]], "xywh")
            image_corners = image_box.corners().astype(float)
            tf_image_from_region = kwimage.Affine.fit(region_corners, image_corners)

            datetime = observable["datetime"]

            visible_polys = []
            for sitesum in site_summaries:
                site_d1 = util_time.coerce_datetime(sitesum["properties"]["start_date"])
                site_d2 = util_time.coerce_datetime(sitesum["properties"]["end_date"])
                # TODO: more date range intersection query, can blend between geom
                # observations
                if site_d1 <= datetime and datetime <= site_d2:
                    wld_site_poly = kwimage.Polygon.coerce(sitesum["geometry"])
                    img_site_poly = wld_site_poly.warp(tf_image_from_region)
                    img_site_poly.meta["color"] = sitesum["properties"]["cache"][
                        "color"
                    ]
                    visible_polys.append(img_site_poly)

            img_width = image_box.width.ravel()[0]
            img_height = image_box.height.ravel()[0]
            image_dsize = [img_width, img_height]
            renderable = {
                "image_dsize": image_dsize,
                "visible_polys": visible_polys,
                "date": datetime,
                "wld_polygon": observable["wld_polygon"],
                "sensor": observable["sensor_name"],
                "frame_idx": frame_idx,
            }
            renderables.append(renderable)

    return region, sites, renderables


def random_observables(num_observations, start_time=None, end_time=None, rng=None):
    """
    Create a random sequence of sensor observations

    Args:
        num_observations (int): number of observations
        start_time (Any): min time coercable
        end_time (Any): max time coercable
        rng : random seed or generator

    Returns:
        List[dict]: list of each item corresonding to a simulated observable

    Example:
        >>> # xdoctest: +SKIP("failing on CI. unsure why")
        >>> from kwgis.demo.metrics_demo.demo_truth import *  # NOQA
        >>> num_observations = 2
        >>> observables = random_observables(1, rng=32)
        >>> print('observables = {}'.format(ub.urepr(observables, nl=2)))
        observables = [
            {
                'datetime': datetime.datetime(2018, 8, 3, 16, 55, 35, 398921),
                'mgrs_code': None,
                'sensor_name': 'demosat-2',
                'source': 'demosat-220180803T165535',
                'wld_polygon': None,
            },
        ]
    """
    rng = kwarray.ensure_rng(rng)
    observed_times = demo_utils.random_time_sequence(
        start_time or "2010-01-01",
        end_time or "2020-01-01",
        num_observations, rng=rng
    )
    # A list of simulated sensors we pretend an observation might be from
    demo_sensors = [
        'demosat-1',
        'demosat-2',
    ]

    # Determine how we made each observation
    observables = []
    sensor_name = "demosat"  # could be more complex
    for datetime in observed_times:
        sensor_name = rng.choice(demo_sensors)
        observables.append(
            {
                "datetime": datetime,
                "sensor_name": sensor_name,
                "source": sensor_name + datetime.strftime('%Y%m%dT%H%M%S'),
                "mgrs_code": None,  # set later
                "wld_polygon": None,  # set later
            }
        )
    return observables


def random_site_model(region_id, site_id, region_corners, observables,
                      site_poly=None, p_observe=0.5, p_transition=0.15,
                      rng=None):
    """
    Make a dummy sequence somewhere within a region's observed spacetime grid.

    Args:
        region_id (str): the name of the region we generate this site for

        site_id (str): the name of the site

        region_corners (ndarray):
            corners of the region to embed this site in.

        observables (List[Dict]):
            information about opportunities to generate an observation

        p_observe (float):
            the probability that a site model is annotated in any particular
            observation.

        p_transition (float):
            truth phase transition model. Currently just the probability
            the phase changes on any particular observation.

        site_poly (kwimage.Polygon | shapely.geometry.Polygon | None):
            if specified, force the site to have this geometry.

        rng : random state or seed

    Returns:
        Tuple[Dict, Dict]: site_summary, site

    Example:
        >>> # xdoctest: +SKIP("failing on CI. unsure why")
        >>> from kwgis.demo.metrics_demo.demo_truth import *  # NOQA
        >>> region_id = 'DR_0042'
        >>> site_id = 'DR_0042_9001'
        >>> rng = kwarray.ensure_rng(0)
        >>> region_corners = kwimage.Boxes([[0, 0, 1, 1]], 'xywh').to_ltrb().corners()
        >>> observables = random_observables(1, rng=rng)
        >>> p_observe = 1.0
        >>> site_summary, site = random_site_model(region_id, site_id, region_corners, observables,
        >>>                                        p_observe=p_observe, rng=rng)
        >>> print('site = {}'.format(ub.urepr(site, nl=4, sort=0)))
        site = {
            'type': 'FeatureCollection',
            'features': [
                {
                    'type': 'Feature',
                    'geometry': {
                        'type': 'Polygon',
                        'coordinates': [[[0.759303, 0.749121], [0.717993, 0.763785], [0.719567, 0.797059], [0.751856, 0.810715], [0.776689, 0.799334], [0.779587, 0.762869], [0.759303, 0.749121]]],
                    },
                    'properties': {
                        'type': 'site',
                        'status': 'positive_annotated',
                        'version': '2.0.1',
                        'site_id': 'DR_0042_9001',
                        'mgrs': None,
                        'start_date': '2015-06-28',
                        'end_date': '2015-06-28',
                        'score': 1,
                        'originator': 'demo',
                        'model_content': 'annotation',
                        'validated': 'True',
                        'misc_info': {'color': [0.0, 1.0, 0.0]},
                        'region_id': 'DR_0042',
                    },
                },
                {
                    'type': 'Feature',
                    'geometry': {
                        'type': 'MultiPolygon',
                        'coordinates': [[[[0.719567, 0.797059], [0.717993, 0.763785], [0.759303, 0.749121], [0.779587, 0.762869], [0.776689, 0.799334], [0.751856, 0.810715]]]],
                    },
                    'properties': {
                        'type': 'observation',
                        'observation_date': '2015-06-28',
                        'source': 'demosat-220150628T072421',
                        'sensor_name': 'demosat-2',
                        'current_phase': 'No Activity',
                        'is_occluded': 'False',
                        'is_site_boundary': 'True',
                        'score': 1.0,
                    },
                },
            ],
        }


    Example:
        >>> from kwgis.demo.metrics_demo.demo_truth import *  # NOQA
        >>> region_id = 'DR_0042'
        >>> site_id = 'DR_0042_9001'
        >>> rng = kwarray.ensure_rng(0)
        >>> region_corners = kwimage.Boxes([[0, 0, 1, 1]], 'xywh').to_ltrb().corners()
        >>> observables = random_observables(10, rng=rng)
        >>> p_observe = 1.0
        >>> site_summary, site = random_site_model(region_id, site_id, region_corners, observables,
        >>>                                        p_observe=p_observe, rng=rng)
        >>> print('site_summary = {}'.format(ub.urepr(site_summary, nl=-1, sort=0)))
        site_summary = {
            'type': 'Feature',
            'geometry': {
                'type': 'Polygon',
                'coordinates': [
                    [
                        [0.599483, 0.568633],
                        [0.569207, 0.576066],
                        [0.542041, 0.627691],
                        [0.606987, 0.67388],
                        [0.647289, 0.636873],
                        [0.635087, 0.580387],
                        [0.599483, 0.568633]
                    ]
                ]
            },
            'properties': {
                'type': 'site_summary',
                'status': 'positive_annotated',
                'version': '2.0.1',
                'site_id': 'DR_0042_9001',
                'mgrs': None,
                'start_date': '2013-11-01',
                'end_date': '2019-08-21',
                'score': 1,
                'originator': 'demo',
                'model_content': 'annotation',
                'validated': 'True',
                'cache': {
                    'color': [1.0, 0.36777954425013254, 0.0]
                }
            }
        }

    Example:
        >>> from kwgis.demo.metrics_demo.demo_truth import *  # NOQA
        >>> region_id = 'DR_0042'
        >>> site_id = 'DR_0042_9001'
        >>> rng = kwarray.ensure_rng(42232)
        >>> region_corners = kwimage.Boxes([[5, 7, 11, 13]], 'xywh').to_ltrb().corners()
        >>> observables = random_observables(10, rng=rng)
        >>> p_observe = 1.0
        >>> site_summary, site = random_site_model(region_id, site_id, region_corners, observables,
        >>>                                        p_observe=p_observe, rng=rng)
        >>> # xdoctest: +REQUIRES(--show)
        >>> # xdoctest: +REQUIRES(module:kwplot)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> import geopandas as gpd
        >>> # Draw our true and perterbed site model
        >>> site_gdf = gpd.GeoDataFrame.from_features(site[1:])
        >>> total_poly = site_gdf['geometry'].unary_union
        >>> minx, miny, maxx, maxy = total_poly.bounds
        >>> fig = kwplot.figure(doclf=True, fnum=1)
        >>> pnum_ = kwplot.PlotNums(nSubplots=len(site_gdf))
        >>> for idx in range(len(site_gdf)):
        >>>     row = site_gdf.iloc[idx: idx + 1]
        >>>     item = row.iloc[0]
        >>>     title = item['observation_date'] + ' ' + item['current_phase']
        >>>     fig = kwplot.figure(pnum=pnum_(), title=title)
        >>>     ax = fig.gca()
        >>>     row.plot(ax=ax, alpha=0.5, color='limegreen', edgecolor='black')
        >>>     row.centroid.plot(ax=ax, alpha=0.5, color='green')
        >>>     ax.set_xlim(minx, maxx)
        >>>     ax.set_ylim(miny, maxy)
        >>> kwplot.show_if_requested()
    """
    rng = kwarray.ensure_rng(rng)

    # Take a subset of the dates we observed this annotation
    n_observables = len(observables)
    observe_flags = rng.rand(n_observables) <= p_observe
    n_tries = 0
    while observe_flags.sum() < min(n_observables, 2):
        # Make sure we have 2 of them
        observe_flags = rng.rand(n_observables) <= p_observe
        n_tries += 1
        if n_tries > 1000:
            raise Exception('Cannot satisfy constraint')

    sampled_observables = list(ub.compress(observables, observe_flags))

    phases = SiteModelGenerator.SITE_PHASES

    # Assign a phase to each date
    curr_phase_idx = 0
    sampled_phases = []
    for _ in range(len(sampled_observables)):
        # could do a more complex transition model
        if curr_phase_idx >= len(phases):
            sampled_observables = sampled_observables[: len(sampled_phases)]
            break
        phase = phases[curr_phase_idx]
        sampled_phases.append(phase)
        if rng.rand() < p_transition:
            curr_phase_idx += 1

    obs_property_defaults = ub.udict(
        {
            "type": "observation",
            "observation_date": None,
            "source": None,
            "sensor_name": None,
            "current_phase": None,
            "is_occluded": "False",  # None,
            "is_site_boundary": "True",  # None,
            "score": 1.0,
        }
    )

    ### Toydata generation parameters
    if site_poly is None:
        # Factors about how the geometry evolves
        max_scale_change = 1.05
        scale_distri = TruncNormal(
            1.0, 0.3, 1 / max_scale_change, max_scale_change, rng=rng
        )

        # Factors into initial size of geometry
        max_scale_factor = 1 / 6
        max_translate = 1 - max_scale_factor

        # Generate a random site polygon in the unit 0-1 quadrent
        # Scale it to ensure its takes at most max_scale_factor of the space
        # And randomly translate within that space.
        offset = rng.rand(2) * max_translate
        site_unit_geom = kwimage.Polygon.random(rng=rng)
        site_unit_geom = site_unit_geom.scale(max_scale_factor).translate(offset)
        sampled_unit_geoms = []
        curr_unit_geom = site_unit_geom.copy()
        for _ in range(len(sampled_observables)):
            # randomly evolve the geometry over time
            scale = scale_distri.sample()
            curr_unit_geom = curr_unit_geom.scale(scale, about='centroid')
            sampled_unit_geoms.append(curr_unit_geom)
        # Build transform from unit to region space
        unit_corners = kwimage.Boxes([[0, 0, 1.0, 1.0]], "xywh").corners()
        tf_region_from_unit = kwimage.Affine.fit(unit_corners, region_corners)

        del offset, max_translate, max_scale_factor, max_scale_change
        del unit_corners
    else:
        site_poly = kwimage.MultiPolygon.coerce(site_poly)
        site_unit_geom = None
        scale_distri = None
        sampled_unit_geoms = [None] * len(sampled_observables)
        tf_region_from_unit = None

    observations = []
    for unit_geom, observable, phase in zip(
        sampled_unit_geoms, sampled_observables, sampled_phases
    ):
        # Warp it into the region space
        datetime = observable["datetime"]
        if site_poly is None:
            site_geom = unit_geom.warp(tf_region_from_unit)
        else:
            site_geom = site_poly
        observations.append(
            geojson.Feature(
                geometry=site_geom.to_multi_polygon().to_geojson(),
                properties=obs_property_defaults | {
                    "observation_date": datetime.date().isoformat(),
                    "current_phase": phase,
                    "sensor_name": observable["sensor_name"],
                    "source": observable["source"],
                },
            )
        )

    if site_poly is None:
        summary_unit_geom = unary_union([g.to_shapely() for g in sampled_unit_geoms]).convex_hull
        summary_unit_poly = kwimage.Polygon.coerce(summary_unit_geom)
        summary_geom = summary_unit_poly.warp(tf_region_from_unit).to_shapely()
    else:
        summary_geom = unary_union(site_poly.to_shapely().geoms)
        if summary_geom.geom_type == 'MultiPolygon':
            summary_geom = summary_geom.convex_hull

    # Build site summary
    status = "positive_annotated"  # Could be more complex here

    mgrs_code = observables[0]["mgrs_code"]

    colors = kwimage.Color.distinct(10)
    idx = rng.randint(0, len(colors) - 1)
    color = colors[idx]

    site_summary = make_site_summary(
        observations, mgrs_code, site_id, status, summary_geom
    )
    if "cache" not in site_summary["properties"]:
        site_summary["properties"]["cache"] = {}
    site_summary["properties"]["cache"]["color"] = color

    site_header = site_summary.copy()
    site_header["properties"] = site_header["properties"].copy()
    site_header["properties"]["type"] = "site"
    site_header["properties"]["region_id"] = region_id
    site = geojson.FeatureCollection([site_header] + observations)
    return site_summary, site


def make_site_summary(observations, mgrs_code, site_id, status, summary_geom=None):
    """
    Consolodate site observations into a site summary
    """
    if summary_geom is None:
        summary_geom = unary_union(
            [kwimage.MultiPolygon.coerce(o["geometry"]).to_shapely() for o in observations]
        ).convex_hull
    if len(observations) == 0:
        raise IndexError('No observations, cannot make a site summary')
    start_date = observations[0]["properties"]["observation_date"]
    end_date = observations[-1]["properties"]["observation_date"]
    sitesum_props = {
        "type": "site_summary",
        "status": status,
        "version": "2.0.1",
        "site_id": site_id,
        "mgrs": mgrs_code,
        "start_date": start_date,
        "end_date": end_date,
        "score": 1,
        "originator": "demo",
        "model_content": "annotation",
        "validated": "True",
    }
    site_summary = geojson.Feature(
        properties=sitesum_props,
        geometry=kwimage.Polygon.coerce(summary_geom).to_geojson(),
    )
    return site_summary



"""
Tools for accessing querying for elevation data.
"""
import numbers
import numpy as np
import os
import random
import requests
import time
import ubelt as ub
from os.path import join


class ElevationDatabase:
    """
    An object that might use various backends to query elevation for a given
    latitude and longitude.

    """

    def __init__(self):
        pass

    @classmethod
    def coerce(cls, key):
        """
        Attempt to resolve key to an elevation database.

        Args:
            key (str): Available options are:
                * "open-elevation"
                * "gtop30"
                * A number for a constant elevation
        """
        if key is False:
            self = ConstantElevationDatabase(0)
        elif isinstance(key, numbers.Number):
            self = ConstantElevationDatabase(key)
        elif key == 'open-elevation':
            self = OpenElevationDatabase()
        elif key == 'gtop30':
            self = girder_gtop30_elevation_dem()
        else:
            raise KeyError(key)
        return self


class ConstantElevationDatabase(ElevationDatabase):
    """
    Fallback compatibility API when no elevation information is available
    """

    def __init__(self, const):
        self.const = const

    def query(self, lats, lons):
        lats_, lons_, was_iterable = ensure_iterable_latlons(lats, lons)
        if was_iterable:
            return np.array([self.const] * len(lats_))
        else:
            return self.const


class OpenElevationDatabase(ElevationDatabase):
    """
    Use open-elevation to query the elevation for a lat/lon point.

    This issues a web request, so it can be slow.

    Args:
        lat (float): degrees in latitude
        lon (float): degrees in longitude
        cache (bool): if True uses on-disk caching
        attempts (int): number of attempts before giving up
        verbose (int): verbosity flag

    Returns:
        float : elevation in meters

    References:
        https://gis.stackexchange.com/questions/338392/getting-elevation-for-multiple-lat-long-coordinates-in-python
        https://gis.stackexchange.com/questions/212106/seeking-alternative-to-google-maps-elevation-api
        https://open-elevation.com/
        https://www.freemaptools.com/elevation-finder.htm

    Example:
        >>> # xdoctest: +REQUIRES(--network)
        >>> # xdoctest: +REQUIRES(--slow)
        >>> from kwgis.gis.elevation import *  # NOQA
        >>> lat = 37.65026538818887
        >>> lon = 128.81096081618637
        >>> eldb = OpenElevationDatabase()
        >>> elevation = eldb.query(lat, lon, verbose=3)
        >>> print('elevation = {!r}'.format(elevation))
        elevation = 449
    """

    def query(self, lat, lon, **kwargs):
        # We could vectorize this, but if is very slow, and we probably should
        # always use gtop30 instead.
        return _query_open_elevation(lat, lon, **kwargs)


def _query_open_elevation(lat, lon, cache=True, attempts=10, verbose=0):
    url = 'https://api.open-elevation.com/api/v1/lookup?'
    suffix = 'locations={},{}'.format(float(lat), float(lon))
    query_url = url + suffix

    cacher = ub.Cacher('elevation', depends=query_url,
                       appname='kwgis/elevation_query', verbose=verbose)
    body = cacher.tryload()
    if body is None:
        for _i in range(attempts):
            result = requests.get(query_url)
            if result.status_code != 200:
                if verbose:
                    print('REQUEST FAILED')
                    print(result.text)
                    print('RETRY')
                time.sleep(3 + random.random() * 3)
            else:
                body = result.json()
                break
        if body is None:
            raise Exception('Failed to query')
        cacher.save(body)
    elevation = body['results'][0]['elevation']
    return elevation


class DEM_Collection(ElevationDatabase):
    """
    Manage a collection of DEM geotiffs

    Example:
        >>> # xdoctest: +REQUIRES(--network)
        >>> # xdoctest: +REQUIRES(--slow)
        >>> # Use the gtop30 DEM dataset from GIRDER to find elevation.
        >>> from kwgis.gis.elevation import *  # NOQA
        >>> dems = DEM_Collection.gtop30()
        >>> lat, lon = (37.7455555555556, 128.780555555556)
        >>> print(dems.query(lat, lon))
        499.0

        >>> # Running one query on the CI is good enough
        >>> # xdoctest: +REQUIRES(--slow)
        >>> # xdoctest: +REQUIRES(--slow)
        >>> lat, lon = (37.7280555555556, 129.008888888889)
        >>> print(dems.query(lat, lon))
        0
        >>> lat, lon = (37.6947222222222, 129.008888888889)
        >>> print(dems.query(lat, lon))
        95.0
        >>> lat, lon = (37.7127777777778, 128.780555555556)
        >>> print(dems.query(lat, lon))
        452.0
        >>> lat, lon = (0, 0)
        >>> print(dems.query(lat, lon))
        0
        >>> lon_basis = np.linspace(-175, 175, 100)
        >>> lat_basis = np.linspace(-85, 85, 100)
        >>> lats_, lons_ = np.meshgrid(lat_basis, lon_basis)
        >>> lats = lats_.ravel()
        >>> lons = lons_.ravel()
        >>> elevations = dems.query(lats, lons)
    """
    def __init__(dems, dem_paths):
        from kwgis.gis.geotiff import geotiff_crs_info
        from kwgis.utils import util_gis
        dem_infos = []
        for dem_fpath in dem_paths:
            dem_info = geotiff_crs_info(dem_fpath)
            dem_infos.append(dem_info)

        dem_crs84_polys = []
        for info in dem_infos:
            gdf = util_gis.crs_geojson_to_gdf(info['geos_corners'])
            gdf_crs84 = gdf.to_crs(util_gis.get_crs84())
            sh_poly = gdf_crs84['geometry'].iloc[0]
            dem_crs84_polys.append(sh_poly)

        dems.dem_paths = dem_paths
        dems.dem_infos = dem_infos
        dems.dem_crs84_polys = dem_crs84_polys

    def __reduce__(self):
        """
        Make this object pickleable by saving references, and then reconstruct
        infos on unpickle. Note: this is inefficient and could be improved.

        Example:
            >>> # xdoctest: +REQUIRES(--network)
            >>> # xdoctest: +REQUIRES(--slow)
            >>> # Check that we can pickle this object
            >>> from kwgis.gis.elevation import *  # NOQA
            >>> dems = DEM_Collection.gtop30()
            >>> import pickle
            >>> text = pickle.dumps(dems)
            >>> recon = pickle.loads(text)
        """
        cls = self.__class__
        args = (self.dem_paths,)
        return (cls, args)

    def find_reference_fpath(dems, lat, lon):
        import shapely

        crs84_query = shapely.geometry.Point(lon, lat)
        flags = [poly.contains(crs84_query) for poly in dems.dem_crs84_polys]
        idxs = np.where(flags)[0]
        assert len(idxs) == 1

        idx = idxs[0]
        dem_fpath = dems.dem_paths[idx]
        dem_info = dems.dem_infos[idx]
        return dem_fpath, dem_info

    def query(dems, lats, lons):
        """
        TODO: the API supports vectorization, but we need to make it more
        efficient.
        """
        from osgeo import gdal
        import kwimage

        lats_, lons_, was_iterable = ensure_iterable_latlons(lats, lons)

        elevations = []

        prev_dem_fpath = None
        prev_dem_band = None

        for lat_, lon_ in zip(lats_, lons_):
            dem_fpath, dem_info = dems.find_reference_fpath(lat_, lon_)

            latlon = kwimage.Coords(np.array([[lat_, lon_]], dtype=np.float64))
            xy = latlon.warp(dem_info['wgs84_to_wld']).warp(dem_info['wld_to_pxl'])

            x, y = np.floor(xy.data[0])
            gdalkw = dict(xoff=x, yoff=y, win_xsize=1, win_ysize=1)

            if prev_dem_fpath != dem_fpath:
                dem_ref = gdal.Open(dem_fpath, gdal.GA_ReadOnly)
                dem_band = dem_ref.GetRasterBand(1)
                prev_dem_band = dem_band
            else:
                dem_band = prev_dem_band

            nodata = dem_band.GetNoDataValue()

            data = dem_band.ReadAsArray(**gdalkw)
            data = data.ravel()
            assert len(data) == 1
            elevation = float(data[0])
            if elevation == nodata:
                # Assume when there is no data
                elevation = 0
            elevations.append(elevation)

        if was_iterable:
            elevations = np.array(elevations)
            return elevations
        else:
            elevation = elevations[0]
            return elevation

    @classmethod
    def gtop30(cls):
        """
        Build the gtop30 dataset

        Example:
            >>> # xdoctest: +REQUIRES(--network)
            >>> # xdoctest: +REQUIRES(--slow)
            >>> dems = DEM_Collection.gtop30()
            >>> lats = [1, 2, 3]
            >>> lons = [1, 2, 3]
            >>> dems.query(lats, lons)
        """
        gtop_dpath = ensure_girder_gtop30_elevation_maps()

        dem_paths = []
        for r, ds, fs in os.walk(gtop_dpath):
            for f in fs:
                dem_paths.append(join(r, f))

        dems = DEM_Collection(dem_paths)
        return dems


def ensure_iterable_latlons(lats, lons):
    lat_was_iterable = ub.iterable(lats)
    lon_was_iterable = ub.iterable(lons)
    was_iterable = lon_was_iterable or lat_was_iterable

    if was_iterable:
        if lon_was_iterable and not lat_was_iterable:
            lons_ = lons
            lats_ = [lats] * len(lons)
        elif lat_was_iterable and not lon_was_iterable:
            lats_ = lats
            lons_ = [lons] * len(lats)
        else:
            lats_ = lats
            lons_ = lons
    else:
        lats_ = [lats]
        lons_ = [lons]

    return lats_, lons_, was_iterable


@ub.memoize
def girder_gtop30_elevation_dem():
    dems = DEM_Collection.gtop30()
    return dems


@ub.memoize
def ensure_girder_gtop30_elevation_maps():
    """
    Ensure that we have the GTOP30 Digital Elevation Maps (DEMS) available
    locally.

    References:
        https://data.kitware.com/#collection/59eb64168d777f31ac6477e7/folder/59fb784d8d777f31ac6480fb
        https://www.google.com/url?sa=j&url=https%3A%2F%2Fwww.usgs.gov%2Fcenters%2Feros%2Fscience%2Fusgs-eros-archive-digital-elevation-global-30-arc-second-elevation-gtopo30%3Fqt-science_center_objects%3D0%23qt-science_center_objects&uct=1599876275&usg=jBvv8w64RCBJd2SyQA3kUtKhMQ4.&source=chat
    """
    print('Building elevation map')
    from kwgis.utils import util_girder
    api_url = 'https://data.kitware.com/api/v1'
    resource_id = '59fb784d8d777f31ac6480fb'
    dl_path = util_girder.grabdata_girder(api_url, resource_id)
    return dl_path



"""
Logic for rendering toy images for generated demo-data.
"""
import kwimage
import kwarray
import numpy as np
from osgeo import osr


def render_toy_georeferenced_image(img_dpath, renderable, rng=None):
    """
    Args:
        img_dpath (PathLike):
            path where the images can be written

        renderable (Dict):
            information about how to render a demo image

        rng : random state

    Returns:
        PathLike:
            The filepath where the image was written

     FIXME:
         I might be doing something wrong with warping polygons from image to
         world space, the y-axis is flipped. Tracking the issue here:
         https://github.com/rasterio/rasterio/issues/2565

    Example:
        >>> from kwgis.demo.metrics_demo.demo_rendering import *  # NOQA
        >>> from kwgis.demo.metrics_demo import demo_utils
        >>> import tempfile
        >>> import kwarray
        >>> from datetime import datetime as datetime_cls
        >>> import ubelt as ub
        >>> img_dpath = ub.Path(tempfile.mkdtemp())
        >>> rng = kwarray.ensure_rng(4324)
        >>> region_poly_wld = demo_utils.random_geo_polygon(rng=rng)
        >>> site_poly_wld = region_poly_wld.scale(0.5, about='center')
        >>> wld_polygon = region_poly_wld
        >>> img_width, img_height = 200, 300
        >>> image_box = kwimage.Boxes([[0, 0, img_width, img_height]], "xywh")
        >>> image_corners = image_box.corners().astype(float)
        >>> wld_box = wld_polygon.bounding_box()
        >>> wld_corners = wld_box.corners()
        >>> tf_img_from_wld = kwimage.Affine.fit(wld_corners, image_corners)
        >>> tf_wld_from_img = tf_img_from_wld.inv()
        >>> site_poly_img = site_poly_wld.warp(tf_img_from_wld)
        >>> renderable = {
        >>>     'sensor': 'foobar',
        >>>     'date': datetime_cls.now(),
        >>>     'frame_idx': 0,
        >>>     'image_dsize': (img_width, img_height),
        >>>     'visible_polys': [site_poly_img],
        >>>     'wld_polygon': wld_polygon,
        >>> }
        >>> fpath = render_toy_georeferenced_image(img_dpath, renderable, rng)
        >>> # xdoctest: +REQUIRES(--show)
        >>> # xdoctest: +REQUIRES(module:kwplot)
        >>> import geopandas as gpd
        >>> gdf = gpd.GeoDataFrame(geometry=[region_poly_wld.to_shapely(),
        >>>                                  site_poly_wld.to_shapely()], crs='OGC:CRS84')
        >>> import kwplot
        >>> kwplot.autompl()
        >>> imdata = kwimage.imread(fpath)
        >>> SHOW_RASTERIO = 1
        >>> if SHOW_RASTERIO:
        >>>     kwplot.imshow(imdata, pnum=(1, 2, 1))
        >>> else:
        >>>     kwplot.imshow(imdata, pnum=(1, 1, 1))
        >>> site_poly_img.draw(edgecolor='black', fill=False)
        >>> # Rasterio rendering has a bug here
        >>> # https://github.com/rasterio/rasterio/issues/2565
        >>> if SHOW_RASTERIO:
        >>>     ax = kwplot.figure(pnum=(1, 2, 2)).gca()
        >>>     import rasterio
        >>>     from rasterio import plot  # NOQA
        >>>     f = rasterio.open(fpath)
        >>>     with rasterio.open(fpath, crs='OGC:CRS84') as f:
        >>>         rasterio.plot.show(f, ax=ax, alpha=0.8)
        >>>         dst_crs = str(f.crs).lower()
        >>>     site_poly_wld.draw(edgecolor='black', fill=0)
        >>> gdf.to_crs(dst_crs).boundary.plot(ax=ax, color='green')
        >>> kwplot.show_if_requested()
    """
    rng = kwarray.ensure_rng(rng)

    sensor = renderable["sensor"]
    frame_idx = renderable["frame_idx"]
    yymmdd = renderable["date"].strftime("%Y%m%d")

    # This is the type of format that can be visualized
    fname = f"frame_{frame_idx}_{yymmdd}_{sensor}.jp2"
    img_fpath = img_dpath / fname

    image_dsize = renderable["image_dsize"]
    wld_polygon = renderable["wld_polygon"]

    # Make dummy image data
    _imdata = kwimage.grab_test_image("amazon")
    imdata = kwimage.imresize(_imdata, dsize=image_dsize)
    imdata = kwimage.ensure_float01(imdata)
    img_h, img_w = imdata.shape[0:2]

    # Speckle noise
    imdata += rng.rand(*imdata.shape) * 0.1
    imdata = imdata.clip(0, 1)

    # Render the polygons on the image
    visible_polys = renderable["visible_polys"]
    for poly in visible_polys:
        color = poly.meta.get("color", None)
        if color is None:
            color = kwimage.Color.random().as01()
        alpha = (rng.rand() * 0.4) + 0.6  # vary alpha
        imdata = poly.draw_on(imdata, color=color, alpha=alpha)

    if 0:
        # More realistic
        nodata_value = -9999
        imdata = (imdata * 10000).astype(np.int16)
    else:
        # Easier to work with
        nodata_value = 0
        imdata = kwimage.ensure_uint255(imdata)

    # Set a region to be nodata_value
    imdata[-10:, 10:] = nodata_value
    imdata[0:10:, -200:-180] = nodata_value
    imdata = imdata[:, :, 0:3]

    write_demo_geotiff(img_fpath=img_fpath, imdata=imdata,
                       wld_polygon=wld_polygon, nodata_value=nodata_value)
    return img_fpath


def write_demo_geotiff(img_fpath=None, imdata=None, wld_polygon=None,
                       nodata_value=None, rng=None, metadata=None):
    """
    Create a demo geotiff at a specified path. Arguments that are not specified
    will be randomly generated.

    Example:
        >>> from kwgis.demo.metrics_demo.demo_rendering import *  # NOQA
        >>> img_fpath = write_demo_geotiff()

    """
    rng = kwarray.ensure_rng(rng)

    if img_fpath is None:
        import ubelt as ub
        import tempfile
        dpath = ub.Path.appdir('kwgis/tests/geotiff/demo').ensuredir()
        img_fpath = tempfile.mktemp(dir=dpath, prefix='demo_geotiff_', suffix='.tif')

    # Generate unspecified data
    if imdata is None:
        imdata = {}

    if isinstance(imdata, dict):
        img_w = imdata.get('width', None)
        img_h = imdata.get('height', None)
        image_dsize = img_w, img_h
        if img_w is None and img_h is None:
            image_dsize = None
        imdata = kwimage.grab_test_image("amazon", dsize=image_dsize)
        imdata = kwimage.ensure_float01(imdata)
        imdata += rng.rand(*imdata.shape) * 0.1
        imdata = imdata.clip(0, 1)
        img_h, img_w = imdata.shape[0:2]

    if metadata is None:
        metadata = {
            'SENSOR': 'kwimage-demo',
            'arbitrary': 'some arbitrary metadata',
        }

    if wld_polygon is None:
        from kwgis.demo.metrics_demo import demo_utils
        wld_polygon = demo_utils.random_geo_polygon(rng=rng)

    # Setup the geo metadata
    wld_box = wld_polygon.bounding_box()
    wld_corners = wld_box.corners()

    img_dsize = imdata.shape[0:2][::-1]
    img_width, img_height = img_dsize

    image_box = kwimage.Boxes([[0, 0, img_width, img_height]], "xywh")
    image_corners = image_box.corners().astype(float)

    # Compute values to trasnform from the world corners to the image corners.
    tf_wld_from_img = kwimage.Affine.fit(image_corners, wld_corners)

    # The CRS should be CRS-84
    srs = osr.SpatialReference()
    srs.ImportFromEPSG(4326)
    srs.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
    crs = srs.ExportToWkt()

    kwimage.imwrite(
        img_fpath,
        imdata,
        backend="gdal",
        nodata_value=nodata_value,
        crs=crs,
        transform=tf_wld_from_img,
        overviews="auto",
        metadata=metadata,
    )
    return img_fpath



r"""
Demodata for a simple region and site model

See Also:
    ../cli/stac_search.py
    ../stac/stac_search_builder.py


CommandLine:

    ###
    # Create the demo dataset
    ###

    # Create a demo region file
    xdoctest -m kwgis.demo.demo_region demo_khq_region_fpath

    DATASET_SUFFIX=DemoKHQ-2022-09-19-V7
    DEMO_DPATH=$HOME/.cache/kwgis/demo/datasets
    REGION_FPATH="$HOME/.cache/kwgis/demo/annotations/KHQ_R001.geojson"
    SITE_GLOBSTR="$HOME/.cache/kwgis/demo/annotations/KHQ_R001_sites/*.geojson"
    REGION_ID=$(jq -r '.features[] | select(.properties.type=="region") | .properties.region_id' "$REGION_FPATH")
    RESULT_FPATH=$DEMO_DPATH/all_sensors_kit/${REGION_ID}.input

    mkdir -p "$DEMO_DPATH"

    # Define SMART_STAC_API_KEY
    source "$HOME"/code/watch/secrets/secrets

    # Delete this to prevent duplicates
    rm -f "$RESULT_FPATH"

    # Construct the TA2-ready dataset
    python -m kwgis.cli.queue_cli.prepare_ta2_dataset \
        --dataset_suffix=$DATASET_SUFFIX \
        --cloud_cover=100 \
        --stac_query_mode=auto \
        --sensors "L2-L8" \
        --api_key=env:SMART_STAC_API_KEY \
        --collated False \
        --requester_pays=True \
        --dvc_dpath="$DEMO_DPATH" \
        --aws_profile=iarpa \
        --region_globstr="$REGION_FPATH" \
        --site_globstr="$SITE_GLOBSTR" \
        --fields_workers=8 \
        --convert_workers=8 \
        --align_workers=26 \
        --cache=0 \
        --ignore_duplicates=1 \
        --separate_region_queues=1 \
        --separate_align_jobs=1 \
        --target_gsd=30 \
        --visualize=True \
        --serial=True --run=1

    # Package up for release on IPFS
    DATASET_DPATH=$DEMO_DPATH/Aligned-$DATASET_SUFFIX

    rm $DATASET_DPATH/img*kwcoco.json
    rm -rf $DATASET_DPATH/_viz512
    rm -rf $DATASET_DPATH/_cache

    7z a $DATASET_DPATH.zip  $DATASET_DPATH

    # Pin the data to IPFS
    DATASET_CID=$(ipfs add -Q -w $DATASET_DPATH.zip --cid-version=1 -s size-1048576)
    echo "On Remote machines run: "
    echo "ipfs pin add $DATASET_CID"

    # Look at the contents of the underlying folder to build scripts.
    echo "DATASET_CID = $DATASET_CID"
    echo "DATASET_SUFFIX=$DATASET_SUFFIX"
    ipfs ls "$DATASET_CID"


    # Pin on a remote service
    ipfs pin remote add --service=web3.storage.erotemic --name="$DATASET_SUFFIX" $DATASET_CID --background
    ipfs pin remote ls --service=web3.storage.erotemic --cid=$DATASET_CID --status=queued,pinning,pinned,failed

    DATASET_CID=bafybeigdkhphpa3n3rdv33w7g6tukmprdnch7g4bp4hc6ebmcr76y6yhwu
    ipfs pin remote ls --service=web3.storage.erotemic --cid=$DATASET_CID --status=queued,pinning,pinned,failed

"""


def demo_khq_annots():
    """
    A small demo region around KHQ while it is being built

    Notes:
        The dates for this demo are rough. THis is the information we have.

            * Some cleanup of the site was happening early in 2017 (and perhaps
              even as early as 2016

            * Land clearing photo 2017-10-06

            * Constructions began in 2017.

            * In late 2017 the first structures went up (the stairwells)

            * Stairwell photo dated 2017-11-19

            * Photo of final building construction stages on 2018-11-29

            * Landscaping continued beyond building construction

            * We moved in in late 2018
    """
    import mgrs
    import kwimage
    import geojson
    import geopandas as gpd
    import ubelt as ub
    from kwgis.utils import util_gis
    from kwutil import util_time

    crs84 = util_gis._get_crs84()

    region_id = 'KHQ_R001'
    site_id = 'KHQ_R001_0000'

    # Boundary of the KHQ construction site
    khq_site_geos = {
        "type": "Polygon",
        "coordinates": [
            [[-73.77200379967688, 42.864783745778894],
             [-73.77177715301514, 42.86412514733195],
             [-73.77110660076141, 42.8641654498268],
             [-73.77105563879013, 42.86423720786224],
             [-73.7710489332676, 42.864399400374786],
             [-73.77134531736374, 42.8649134986743],
             [-73.77200379967688, 42.864783745778894]]
        ]
    }

    sites = []
    if 1:
        # Detailed observation on KHQ construction
        obs_property_defaults = {
            'type': 'observation',
            'observation_date': None,
            'source': None,
            'sensor_name': None,
            'current_phase': None,
            'is_occluded': None,
            'is_site_boundary': None,
            'score': 1.0,
        }

        observations = []
        observations.append(geojson.Feature(
            properties=ub.dict_union(obs_property_defaults, {
                "observation_date": "2016-12-01",
                "current_phase": 'No Activity',
                "source": 'guess',
            }),
            geometry=khq_site_geos,
        ))
        observations.append(geojson.Feature(
            geometry=khq_site_geos,
            properties=ub.dict_union(obs_property_defaults, {
                "observation_date": "2017-09-13",
                "current_phase": 'Site Preparation',
                'sensor_name': 'S2_L2A',
                "source": 'sentinel-2',
            })))
        observations.append(geojson.Feature(
            geometry=khq_site_geos,
            properties=ub.dict_union(obs_property_defaults, {
                "observation_date": "2017-10-03",
                "current_phase": 'Site Preparation',
                'sensor_name': 'S2_L2A',
                "source": 'sentinel-2',
            })))
        observations.append(geojson.Feature(
            geometry=khq_site_geos,
            properties=ub.dict_union(obs_property_defaults, {
                "observation_date": "2017-10-06",
                "current_phase": 'Site Preparation',
                "source": 'ground-photo',
            })))
        observations.append(geojson.Feature(
            geometry=khq_site_geos,
            properties=ub.dict_union(obs_property_defaults, {
                "observation_date": "2017-11-19",
                "current_phase": 'Active Construction',
                "source": 'ground-photo',
            })))
        observations.append(geojson.Feature(
            geometry=khq_site_geos,
            properties=ub.dict_union(obs_property_defaults, {
                "observation_date": "2018-11-29",
                "current_phase": 'Active Construction',
                "source": 'ground-photo',
            })))
        observations.append(geojson.Feature(
            geometry=khq_site_geos,
            properties=ub.dict_union(obs_property_defaults, {
                "observation_date": "2019-01-01",
                "current_phase": 'Post Construction',
                "source": 'guess',
            })))
        obs_df = gpd.GeoDataFrame(observations, crs=crs84)
        # site_boundary_poly = kwimage.Polygon.coerce(obs_df.geometry.unary_union)
        # Use context manager to do all transforms in UTM space
        with util_gis.UTM_TransformContext(obs_df) as context:
            site_boundary_poly = context.geoms_utm.unary_union
            context.finalize(site_boundary_poly)
        site_boundary_geom = context.final_geoms_crs84.iloc[0]
        site_boundary_poly = kwimage.Polygon.coerce(site_boundary_geom)
        site_boundary_geos = site_boundary_poly.to_geojson()

        obs_dates = [
            util_time.coerce_datetime(obs['properties']['observation_date'])
            for obs in observations
        ]
        start_date = min(obs_dates)
        end_date = max(obs_dates)

        lon, lat = kwimage.Polygon.from_geojson(khq_site_geos).centroid
        mgrs_code = mgrs.MGRS().toMGRS(lat, lon, MGRSPrecision=0)
        khq_sitesum = geojson.Feature(
            properties={
                "type": "site",
                "region_id": region_id,
                "site_id": site_id,
                "version": "2.0.0",
                "status": "positive_annotated",
                "mgrs": mgrs_code,
                "score": 1.0,
                "start_date": start_date.date().isoformat(),
                "end_date": end_date.date().isoformat(),
                "model_content": "annotation",
                "originator": "kit-demo",
                "validated": "False",
            },
            geometry=site_boundary_geos
        )
        khq_site = geojson.FeatureCollection(
            [khq_sitesum] + observations
        )
        sites.append(khq_site)

    # Build site summaries for each site (there is only one site)
    site_summaries = []
    for site in sites:
        # The property type changes from site to site summary.
        # Not sure why, but dems da specs
        site_header = site['features'][0]
        site_summary = geojson.Feature(
            properties=ub.dict_union(
                site_header['properties'], {'type': 'site_summary'}),
            geometry=site_header['geometry'],
        )
        site_summaries.append(site_summary)

    # Aggregate information across sites to build info for the region
    site_geoms = []
    site_start_dates = []
    site_end_dates = []
    for site in site_summaries:
        site_geoms.append(
            kwimage.Polygon.coerce(site['geometry']).to_shapely()
        )
        start_date = site['properties']['start_date']
        if start_date is not None:
            site_start_dates.append(util_time.coerce_datetime(start_date))
        end_date = site['properties']['end_date']
        if end_date is not None:
            site_end_dates.append(util_time.coerce_datetime(end_date))

    # Custom region geom beyond that provided by sites
    khq_region_geom = kwimage.Polygon.coerce({
        "type": "Polygon",
        "coordinates": [
            [[-73.77379417419434, 42.86254939745846],
             [-73.76715302467346, 42.86361104246733],
             [-73.76901984214783, 42.86713400027327],
             [-73.77529621124268, 42.865978051904726],
             [-73.7755537033081, 42.86542759269259],
             [-73.7750494480133, 42.862525805139775],
             [-73.77379417419434, 42.86254939745846]]
        ]
    }).to_shapely()
    site_geoms.append(khq_region_geom)

    # Use context manager to do all transforms in UTM space
    with util_gis.UTM_TransformContext(site_geoms) as context:
        tmp_geom_utm = context.geoms_utm.unary_union
        tmp_poly_utm = kwimage.Polygon.coerce(tmp_geom_utm)
        agg_geom_utm = tmp_poly_utm.scale(2.0, about='center').to_shapely()
        context.finalize(agg_geom_utm)
    khq_region_poly = context.final_geoms_crs84.iloc[0]
    khq_region_geom = kwimage.Polygon.coerce(khq_region_poly).to_geojson()

    delta_pad = util_time.coerce_timedelta('1460days')

    khq_region_start_time = min(site_start_dates) - delta_pad
    khq_region_end_time = max(site_end_dates) + delta_pad

    # Enlarge the region
    khq_region_feature = geojson.Feature(
        properties={
            "type": "region",
            "region_id": "KHQ_R001",
            "version": "2.4.3",
            "mgrs": mgrs_code,
            "start_date": khq_region_start_time.date().isoformat(),
            "end_date": khq_region_end_time.date().isoformat(),
            "originator": "kit-demo",
            "model_content": "annotation",
            "comments": 'kwgis-demo-data',
        },
        geometry=khq_region_geom
    )
    region = geojson.FeatureCollection([khq_region_feature] + site_summaries)
    return region, sites


def demo_smart_annots():
    """
    A small demo region in an area with a lot of data coverage
    """
    import mgrs
    import kwimage
    import geojson
    import geopandas as gpd
    import ubelt as ub
    from kwgis.utils import util_gis
    from kwutil import util_time

    crs84 = util_gis._get_crs84()

    region_id = 'SDEMO_R001'
    site_id = 'SDEMO_R001_0000'

    # Boundary of the KHQ construction site
    site_geos = {'type': 'Polygon',
                 'coordinates': [[[-81.70861006, 30.3741017],
                                  [-81.70745055, 30.28932539],
                                  [-81.58419451, 30.29015089],
                                  [-81.58346045, 30.37588478],
                                  [-81.70861006, 30.3741017]]]}

    sites = []
    WITH_SITES = 0
    if WITH_SITES:
        # obs_property_defaults = {
        #     'type': 'observation',
        #     'observation_date': None,
        #     'source': None,
        #     'sensor_name': None,
        #     'current_phase': None,
        #     'is_occluded': None,
        #     'is_site_boundary': None,
        #     'score': 1.0,
        # }

        # No observations in this case
        observations = []
        obs_df = gpd.GeoDataFrame(observations, crs=crs84)
        # site_boundary_poly = kwimage.Polygon.coerce(obs_df.geometry.unary_union)
        # Use context manager to do all transforms in UTM space
        with util_gis.UTM_TransformContext(obs_df) as context:
            site_boundary_poly = context.geoms_utm.unary_union
            context.finalize(site_boundary_poly)
        site_boundary_geom = context.final_geoms_crs84.iloc[0]
        site_boundary_poly = kwimage.Polygon.coerce(site_boundary_geom)
        site_boundary_geos = site_boundary_poly.to_geojson()

        obs_dates = [
            util_time.coerce_datetime(obs['properties']['observation_date'])
            for obs in observations
        ]
        start_date = min(obs_dates)
        end_date = max(obs_dates)

        lon, lat = kwimage.Polygon.from_geojson(site_geos).centroid
        mgrs_code = mgrs.MGRS().toMGRS(lat, lon, MGRSPrecision=0)
        sitesum = geojson.Feature(
            properties={
                "type": "site",
                "region_id": region_id,
                "site_id": site_id,
                "version": "2.0.0",
                "status": "positive_annotated",
                "mgrs": mgrs_code,
                "score": 1.0,
                "start_date": start_date.date().isoformat(),
                "end_date": end_date.date().isoformat(),
                "model_content": "annotation",
                "originator": "kit-demo",
                "validated": "False",
            },
            geometry=site_boundary_geos
        )
        site = geojson.FeatureCollection(
            [sitesum] + observations
        )
        sites.append(site)

    # Build site summaries for each site (there is only one site)
    site_summaries = []
    for site in sites:
        # The property type changes from site to site summary.
        # Not sure why, but dems da specs
        site_header = site['features'][0]
        site_summary = geojson.Feature(
            properties=ub.dict_union(
                site_header['properties'], {'type': 'site_summary'}),
            geometry=site_header['geometry'],
        )
        site_summaries.append(site_summary)

    # Aggregate information across sites to build info for the region
    site_geoms = []
    site_start_dates = []
    site_end_dates = []
    for site in site_summaries:
        site_geoms.append(
            kwimage.Polygon.coerce(site['geometry']).to_shapely()
        )
        start_date = site['properties']['start_date']
        if start_date is not None:
            site_start_dates.append(util_time.coerce_datetime(start_date))
        end_date = site['properties']['end_date']
        if end_date is not None:
            site_end_dates.append(util_time.coerce_datetime(end_date))

    # Custom region geom beyond that provided by sites
    region_geom = kwimage.Polygon.coerce(site_geos).scale(1.5, about='centroid').to_shapely()
    site_geoms.append(region_geom)

    # Use context manager to do all transforms in UTM space
    with util_gis.UTM_TransformContext(site_geoms) as context:
        tmp_geom_utm = context.geoms_utm.unary_union
        tmp_poly_utm = kwimage.Polygon.coerce(tmp_geom_utm)
        agg_geom_utm = tmp_poly_utm.scale(2.0, about='center').to_shapely()
        context.finalize(agg_geom_utm)
    region_poly = context.final_geoms_crs84.iloc[0]
    region_geom = kwimage.Polygon.coerce(region_poly).to_geojson()

    delta_pad = util_time.coerce_timedelta('128days')

    site_start_dates.append(util_time.coerce_datetime('2017-01-01'))
    site_end_dates.append(util_time.coerce_datetime('2017-01-01'))

    region_start_time = min(site_start_dates) - delta_pad
    region_end_time = max(site_end_dates) + delta_pad

    lon, lat = kwimage.Polygon.coerce(region_geom).centroid
    mgrs_code = mgrs.MGRS().toMGRS(lat, lon, MGRSPrecision=0)

    # Enlarge the region
    # TODO: see RegionModel in ~/code/watch/kwgis/cli/cluster_sites.py
    region_feature = geojson.Feature(
        properties={
            "type": "region",
            "region_id": region_id,
            "version": "2.4.3",
            "mgrs": mgrs_code,
            "start_date": region_start_time.date().isoformat(),
            "end_date": region_end_time.date().isoformat(),
            "originator": "kit-demo",
            "model_content": "annotation",
            "comments": 'kwgis-demo-data',
        },
        geometry=region_geom
    )
    region = geojson.FeatureCollection([region_feature] + site_summaries)
    return region, sites


def _configure_osm():
    """
    Configure open street map
    """
    import osmnx as ox
    import ubelt as ub
    import os
    osm_settings_dirs_varnames = [
        'data_folder',
        'logs_folder',
        'imgs_folder',
        'cache_folder',
    ]
    # Make osm dirs point at a standardized location
    osm_cache_root = ub.Path.appdir('osm')
    for varname in osm_settings_dirs_varnames:
        val = ub.Path(getattr(ox.settings, varname))
        if not val.is_absolute():
            new_val = os.fspath(osm_cache_root / os.fspath(val))
            setattr(ox.settings, varname, new_val)

    ox.settings.log_console = True
    ox.settings.log_console = True
    return ox


def _show_demo_annots_on_map():
    """
    Test to check the demo features on a map

    Requires:
        pip install osmnx

    Example:
        >>> # xdoctest: +REQUIRES(--show)
        >>> from kwgis.demo.demo_region import *  # NOQA
        >>> from kwgis.demo.demo_region import _show_demo_annots_on_map
        >>> import kwplot
        >>> kwplot.autompl()
    """
    import io
    import json
    import kwimage
    import kwplot
    import matplotlib
    from kwgis.utils import util_gis
    ox = _configure_osm()
    # https://geopandas.org/en/stable/gallery/plotting_basemap_background.html
    kwplot.autompl()

    region, sites = demo_khq_annots()
    region_gdf = util_gis.load_geojson(io.StringIO(json.dumps(region)))

    region_geom = region_gdf.geometry.iloc[0]
    osm_graph = ox.graph_from_polygon(region_geom)

    ax = kwplot.figure(fnum=1, docla=True).gca()
    fig, ax = ox.plot_graph(osm_graph, bgcolor='lawngreen', node_color='dodgerblue', edge_color='skyblue', ax=ax)
    print(matplotlib.get_backend())

    for geom in region_gdf.geometry:
        kwpoly = kwimage.Polygon.coerce(geom)
        kwpoly.draw(ax=ax, facecolor='none', fill=None, linewidth=1)

    # import contextily as cx
    # import mplcairo  # NOQA
    # cx.add_basemap(ax, url=cx.providers.OpenStreetMap.Mapnik)
    # cx.add_basemap(ax)
    # cx.add_basemap(ax, url=cx.providers.Esri.WorldImagery)


def demo_khq_region_fpath():
    import json
    import ubelt as ub
    region, sites = demo_khq_annots()
    annot_dpath = ub.Path.appdir('kwgis/demo/annotations').ensuredir()

    # Dump region file
    region_id = region['features'][0]['properties']['region_id']
    region_fpath = annot_dpath / (region_id + '.geojson')
    region_fpath.write_text(json.dumps(region))
    print(f'wrote region_fpath={region_fpath}')

    # Dump site file
    site_dpath = (annot_dpath / (region_id + '_sites')).ensuredir()
    for site in sites:
        site_id = site['features'][0]['properties']['site_id']
        site_fpath = site_dpath / (site_id + '.geojson')
        site_fpath.write_text(json.dumps(site))
        print(f'wrote site_fpath={site_fpath}')
    return region_fpath


def demo_smart_region_fpath():
    import json
    import ubelt as ub
    region, sites = demo_smart_annots()
    annot_dpath = ub.Path.appdir('kwgis/demo/annotations').ensuredir()

    # Dump region file
    region_id = region['features'][0]['properties']['region_id']
    region_fpath = annot_dpath / (region_id + '.geojson')
    region_fpath.write_text(json.dumps(region))
    print(f'wrote region_fpath={region_fpath}')

    # Dump site file
    site_dpath = (annot_dpath / (region_id + '_sites')).ensuredir()
    for site in sites:
        site_id = site['features'][0]['properties']['site_id']
        site_fpath = site_dpath / (site_id + '.geojson')
        site_fpath.write_text(json.dumps(site))
        print(f'wrote site_fpath={site_fpath}')
    return region_fpath


if __name__ == '__main__':
    """
    CommandLine:
        python ~/code/watch/kwgis/demo/demo_region.py
    """
    demo_khq_region_fpath()



"""
A tool to help build a STAC search json.

This does contain a command line interface, for legacy reasons, but
is not intended to be a fully supported part of the GeoWATCH CLI.

SeeAlso:
    ../cli/stac_search.py

Accenture Notes:

    New Procesing 2022-11-21:
        https://smart-research.slack.com/?redir=%2Ffiles%2FU028UQGN1N0%2FF04B998ANRL%2Faccenture_ta1_productdoc_phaseii_20211117.pptx%3Forigin_team%3DTN3QR7WAH%26origin_channel%3DC03QTAXU7GF

        https://smartgitlab.com/TE/evaluations/-/wikis/Accenture-TA-1-Processing-Status
"""
import os
import ubelt as ub
import scriptconfig as scfg


class StacSearchBuilderConfig(scfg.DataConfig):
    """
    Helper to create STAC search json queries
    """
    __default__ = {
        'start_date': scfg.Value(None, help='iso starting date'),
        'end_date': scfg.Value(None, help='iso ending date'),
        'cloud_cover': scfg.Value(10, help='maximum cloud cover percentage'),
        'out_fpath': scfg.Value(None, help='if unspecified, write to stdout'),
        'api_key': scfg.Value('env:SMART_STAC_API_KEY', help='The API key or where to get it'),
        'sensors': scfg.Value('L2-S2', help=''),
    }


_KITWARE_PHASE1_TA1_PRODUCTS = {
    # Kitware Phase 1 TA-1 Products
    'ta1-s2-kit': {
        'endpoint': "https://api.smart-stac.com",
        'collections': ['ta1-s2-kit'],
    },
    'ta1-l8-kit': {
        'endpoint': "https://api.smart-stac.com",
        'collections': ['ta1-l8-kit'],
    },
    'ta1-wv-kit': {
        'endpoint': "https://api.smart-stac.com",
        'collections': ['ta1-wv-kit'],
        "query": {
            "nitf:imd": {
                "eq": "true"
            },
        }
    },
}


_ACCENTURE_PHASE2_TA1_PRODUCTS = {
    # Accenture Phase 2 TA-1 Products
    'ta1-s2-acc': {
        'endpoint': "https://api.smart-stac.com",
        'collections': ['ta1-s2-acc'],
    },
    'ta1-ls-acc': {
        'endpoint': "https://api.smart-stac.com",
        'collections': ['ta1-ls-acc'],
    },
    'ta1-pd-acc': {
        'endpoint': "https://api.smart-stac.com",
        'collections': ['ta1-pd-acc'],
    },
    'ta1-wv-acc': {
        'endpoint': "https://api.smart-stac.com",
        'collections': ['ta1-wv-acc'],
        "query": {
            "nitf:imd": {
                "eq": "true"
            },
        }
    },

    # Accenture Phase 2 TA-1 Products
    'ta1-s2-acc-1': {
        'endpoint': "https://api.smart-stac.com",
        'collections': ['ta1-s2-acc-1'],
    },
    'ta1-ls-acc-1': {
        'endpoint': "https://api.smart-stac.com",
        'collections': ['ta1-ls-acc-1'],
    },
    'ta1-pd-acc-1': {
        'endpoint': "https://api.smart-stac.com",
        'collections': ['ta1-pd-acc-1'],
    },
    'ta1-wv-acc-1': {
        'endpoint': "https://api.smart-stac.com",
        'collections': ['ta1-wv-acc-1'],
        "query": {
            "nitf:imd": {
                "eq": "true"
            },
        }
    },

    'ta1-s2-acc-2': {
        'endpoint': "https://api.smart-stac.com",
        'collections': ['ta1-s2-acc-2'],
    },
    'ta1-ls-acc-2': {
        'endpoint': "https://api.smart-stac.com",
        'collections': ['ta1-ls-acc-2'],
    },
    'ta1-pd-acc-2': {
        'endpoint': "https://api.smart-stac.com",
        'collections': ['ta1-pd-acc-2'],
    },
    'ta1-wv-acc-2': {
        'endpoint': "https://api.smart-stac.com",
        'collections': ['ta1-wv-acc-2'],
        "query": {
            "nitf:imd": {
                "eq": "true"
            },
        }
    },

    'ta1-s2-acc-3': {
        'endpoint': "https://api.smart-stac.com",
        'collections': ['ta1-s2-acc-3'],
    },
    'ta1-ls-acc-3': {
        'endpoint': "https://api.smart-stac.com",
        'collections': ['ta1-ls-acc-3'],
    },
    'ta1-pd-acc-3': {
        'endpoint': "https://api.smart-stac.com",
        'collections': ['ta1-pd-acc-3'],
    },
    'ta1-wv-acc-3': {
        'endpoint': "https://api.smart-stac.com",
        'collections': ['ta1-wv-acc-3'],
        "query": {
            "nitf:imd": {
                "eq": "true"
            },
        }
    },

    'ta1-10m-tsmoothed-acc-3': {
        'endpoint': "https://api.smart-stac.com",
        'collections': ['ta1-10m-tsmoothed-acc-3'],
    }

}

_ARA_PHASE3_TA1_PRODUCTS = {
    'ta1-ls-ara-4': {
        'endpoint': "https://api.smart-stac.com",
        'collections': ['ta1-ls-ara-4'],
    },
    'ta1-pd-ara-4': {
        'endpoint': "https://api.smart-stac.com",
        'collections': ['ta1-pd-ara-4'],
    },
    'ta1-s2-ara-4': {
        'endpoint': "https://api.smart-stac.com",
        'collections': ['ta1-s2-ara-4'],
    },
    'ta1-wv-ara-4': {
        'endpoint': "https://api.smart-stac.com",
        'collections': ['ta1-wv-ara-4'],
        "query": {
            "nitf:imd": {
                "eq": "true"
            },
        }
    },
}


# ta1-mixedgsd-acc


_SMARTSTAC_PRODUCTS = {
    # Non public, but non-performer products
    'planet-dove': {
        'endpoint': "https://api.smart-stac.com",
        'collections': ['planet-dove'],
    },
    'worldview-nitf': {
        'endpoint': "https://api.smart-stac.com",
        'collections': ['worldview-nitf'],
        "query": {
            "nitf:imd": {
                "eq": "true"
            },
        }
    },

    'smart-landsat-c2l1': {
        "collections": ["landsat-c2l1"],
        'endpoint': "https://api.smart-stac.com",
        "query": {
            "platform": {
                "eq": "LANDSAT_8"
            }
        }
    },

    'smart-landsat-c2l2-sr': {
        "collections": ["landsat-c2l2-sr"],
        'endpoint': "https://api.smart-stac.com",
        "query": {
            "platform": {
                "eq": "LANDSAT_8"
            }
        }
    },

}


_PUBLIC_L1_PRODUCTS = {
    # https://landsatlook.usgs.gov/stac-server/
    # Public L1 Products
    'landsat-c2l1': {
        "collections": ["landsat-c2l1"],
        "endpoint": "https://landsatlook.usgs.gov/stac-server/",
        "query": {
            "platform": {
                "eq": "LANDSAT_8"
            }
        }
    },
    'sentinel-2-l1c': {
        "collections": ["sentinel-2-l1c"],
        "endpoint": "https://earth-search.aws.element84.com/v1",
    },
}


# NOTE;
# Info about QA bands:
# https://d9-wret.s3.us-west-2.amazonaws.com/assets/palladium/production/s3fs-public/media/files/LSDS-1435%20Landsat%20C2%20US%20ARD%20Data%20Format%20Control%20Book-v3.pdf

# Updated L2 Products
# https://www.element84.com/geospatial/introducing-earth-search-v1-new-datasets-now-available/

_PUBLIC_L2_PRODUCTS = {
    # Public L2 Products
    # https://docs.sentinel-hub.com/api/latest/data/sentinel-2-l2a/
    'sentinel-2-l2a': {
        "collections": ["sentinel-2-l2a"],
        "endpoint": "https://earth-search.aws.element84.com/v1",
    },

    # https://stacindex.org/catalogs/usgs-landsat-collection-2-api#/
    # Surface Reflectance
    'landsat-c2l2-sr': {
        # Note: AWS_REQUEST_PAYER='requester' is required to grab the data
        "collections": ["landsat-c2l2-sr"],
        "endpoint": "https://landsatlook.usgs.gov/stac-server/",
        "query": {
            "platform": {
                "eq": "LANDSAT_8"
            }
        }
    },

    # https://stacindex.org/catalogs/usgs-landsat-collection-2-api#/
    # Brightness Temperature
    'landsat-c2l2-bt': {
        # Note: AWS_REQUEST_PAYER='requester' is required to grab the data
        "collections": ["landsat-c2l2-bt"],
        "endpoint": "https://landsatlook.usgs.gov/stac-server/",
        "query": {
            "platform": {
                "eq": "LANDSAT_8"
            }
        }
    },
}


# https://www.drivendata.org/competitions/143/tick-tick-bloom/page/650/#sentinel-2-1
# https://planetarycomputer.microsoft.com/dataset/landsat-c2-l2
_PUBLIC_L2_PRODUCTS['planetarycomputer'] = {
    'collections': ['landsat-c2-l2', 'sentinel-2-l2a'],
    "endpoint": "https://planetarycomputer.microsoft.com/api/stac/v1",
}

_PUBLIC_L2_PRODUCTS['planetarycomputer_l8'] = {
    'collections': ['landsat-c2-l2'],
    "endpoint": "https://planetarycomputer.microsoft.com/api/stac/v1",
}

_PUBLIC_L2_PRODUCTS['planetarycomputer_s2'] = {
    'collections': ['sentinel-2-l2a'],
    "endpoint": "https://planetarycomputer.microsoft.com/api/stac/v1",
}


_PUBLIC_ARD_PRODUCTS = {
    # https://stacindex.org/catalogs/usgs-landsat-collection-2-api#/
    # Surface Reflectance
    'landsat-c2ard-sr': {
        # Note: AWS_REQUEST_PAYER='requester' is required to grab the data
        "collections": ["landsat-c2ard-sr"],
        "endpoint": "https://landsatlook.usgs.gov/stac-server/",
        "query": {
            "platform": {
                "eq": "LANDSAT_8"
            }
        }
    },

    # https://stacindex.org/catalogs/usgs-landsat-collection-2-api#/
    # Brightness Temperature
    'landsat-c2ard-bt': {
        # Note: AWS_REQUEST_PAYER='requester' is required to grab the data
        "collections": ["landsat-c2ard-bt"],
        "endpoint": "https://landsatlook.usgs.gov/stac-server/",
        "query": {
            "platform": {
                "eq": "LANDSAT_8"
            }
        }
    },
}


# Note: these functions moved into the _notebook.py file in this directory
def print_provider_debug_information():
    raise NotImplementedError('Moved to the _notebook.py in this folder')


def check_processed_regions():
    raise NotImplementedError('Moved to the _notebook.py in this folder')


def _devcheck_providers_exist():
    raise NotImplementedError('Moved to the _notebook.py in this folder')

### Available smartstac collections:
# <CollectionClient id=landsat-c2l2-sr>,
# <CollectionClient id=ta1-s2-ara>,
# <CollectionClient id=ta1-pd-ara>,
# <CollectionClient id=ssh-wv-acc>,
# <CollectionClient id=ssh-ls-acc>,
# <CollectionClient id=ta1-wv-ara>,
# <CollectionClient id=ta1-ls-ara>,
# <CollectionClient id=ta1-s2-acc>,
# <CollectionClient id=worldview-nitf>,
# <CollectionClient id=ta1-wv-str>,
# <CollectionClient id=landsat-c2l1>,
# <CollectionClient id=ta1-ls-str>,
# <CollectionClient id=ta1-pd-str>,
# <CollectionClient id=ta1-pd-kit>,
# <CollectionClient id=planet-dove>,
# <CollectionClient id=ta1-ls-kit>,
# <CollectionClient id=ta1-wv-acc>,
# <collectionclient id=ta1-dsm-ara>,
# <CollectionClient id=ssh-pd-acc>,
# <CollectionClient id=ssh-s2-acc>,
# <CollectionClient id=ta1-s2-kit>,
# <CollectionClient id=ta1-wv-kit>,
# <CollectionClient id=ta1-pd-acc>,
# <CollectionClient id=ta1-ls-acc>,
# <CollectionClient id=ta1-s2-str>,


SENSOR_TO_DEFAULTS = ub.dict_union(
    _KITWARE_PHASE1_TA1_PRODUCTS,
    _ACCENTURE_PHASE2_TA1_PRODUCTS,
    _PUBLIC_L1_PRODUCTS,
    _PUBLIC_L2_PRODUCTS,
    _PUBLIC_ARD_PRODUCTS,
    _SMARTSTAC_PRODUCTS,
    _ARA_PHASE3_TA1_PRODUCTS,
)


# Simplified codes for the CLI
CONVINIENCE_SENSOR_GROUPS = {
    'TA1': [
        'ta1-s2-kit',
        # 'ta1-l8-kit',
        'ta1-wv-kit',
    ],
    'L2-S2': [
        'sentinel-2-l2a',
    ],
    'L2-L8': [
        'landsat-c2l2-sr',
        'landsat-c2l2-bt',
    ],
    'L2-S2-L8': [
        'sentinel-2-l2a',
        'landsat-c2l2-sr',
        'landsat-c2l2-bt',
    ],
    'ARD-L8': [
        'landsat-c2ard-sr',
        'landsat-c2ard-bt',
    ],
    'ARD-S2-L8': [
        'sentinel-2-l2a',
        'landsat-c2ard-sr',
        'landsat-c2ard-bt',
    ],

    'TA1-S2-L8-ACC': [
        'ta1-s2-acc',
        'ta1-ls-acc',
    ],
    'TA1-S2-L8-WV-PD-ACC-2': [
        'ta1-s2-acc-2',
        'ta1-ls-acc-2',
        'ta1-pd-acc-2',
        'ta1-wv-acc-2',
    ],
    'TA1-S2-L8-WV-PD-ACC-3': [
        'ta1-s2-acc-3',
        'ta1-ls-acc-3',
        'ta1-pd-acc-3',
        'ta1-wv-acc-3',
    ],
    'TA1-S2-L8-WV-PD-ACC-1': [
        'ta1-s2-acc-1',
        'ta1-ls-acc-1',
        'ta1-pd-acc-1',
        'ta1-wv-acc-1',
    ],
    'TA1-S2-L8-WV-PD-ACC': [
        'ta1-s2-acc',
        'ta1-ls-acc',
        'ta1-pd-acc',
        'ta1-wv-acc',
    ],
    'TA1-S2-WV-PD-ACC': [
        'ta1-s2-acc',
        'ta1-pd-acc',
        'ta1-wv-acc',
    ],
    'TA1-WV-PD-ACC': [
        'ta1-pd-acc',
        'ta1-wv-acc',
    ],
    'TA1-S2-ACC': [
        'ta1-s2-acc',
    ],
    'TA1-L8-ACC': [
        'ta1-ls-acc',
    ],
    'TA1-10M-TSMOOTH-ACC-3': [
        'ta1-10m-tsmoothed-acc-3',
    ],
    'TA1-S2-ACC-3': [
        'ta1-s2-acc-3',
    ],
}


def build_search_json(start_date, end_date, sensors, api_key, cloud_cover):
    """
    Construct the json that can be used for a stac search

    Example:
        >>> from kwgis.stac.stac_search_builder import build_search_json
        >>> start_date = '2017-01-01'
        >>> end_date = '2020-01-01'
        >>> sensors = 'L2-S2'
        >>> api_key = None
        >>> cloud_cover = 20
        >>> search_json = build_search_json(start_date, end_date, sensors, api_key, cloud_cover)
        >>> print('search_json = {}'.format(ub.urepr(search_json, nl=-1)))
        search_json = {
            'stac_search': [
                {
                    'collections': ['sentinel-2-l2a'],
                    'endpoint': 'https://earth-search.aws.element84.com/v1',
                    'query': {
                        'eo:cloud_cover': {'lt': 20}
                    },
                    'headers': {},
                    'start_date': '2017-01-01',
                    'end_date': '2020-01-01'
                }
            ]
        }

    """
    from kwutil import util_time

    if api_key is not None and api_key.startswith('env:'):
        api_environ_key = api_key.split(':')[1]
        api_key = os.environ.get(api_environ_key, None)

    if isinstance(sensors, str):
        if sensors in SENSOR_TO_DEFAULTS:
            sensors = [sensors]
        else:
            sensors = CONVINIENCE_SENSOR_GROUPS[sensors]

    headers = {}
    if api_key is not None:
        headers['x-api-key'] = api_key

    start_date = util_time.coerce_datetime(start_date, default_timezone='utc')
    end_date = util_time.coerce_datetime(end_date, default_timezone='utc')

    if start_date is None or end_date is None:
        raise ValueError('need start and end date')

    search_item_list = []
    for sensor in sensors:
        search_item = SENSOR_TO_DEFAULTS[sensor]
        item_query = search_item.setdefault('query', {})
        search_item['headers'] = headers
        search_item['start_date'] = start_date.date().isoformat()
        search_item['end_date'] = end_date.date().isoformat()
        if cloud_cover is not None:
            item_query['eo:cloud_cover'] = {
                'lt': cloud_cover,
            }
        search_item_list.append(search_item)

    search_json = {
        'stac_search': search_item_list,
    }
    return search_json


def main(cmdline=1, **kwargs):
    """
    Example:
        >>> # xdoctest: +SKIP
        >>> # xdoctest: +REQUIRES(env:SLOW_DOCTEST)
        >>> from kwgis.stac.stac_search_builder import main
        >>> cmdline = 0
        >>> kwargs = {
        >>>     'start_date': '2017-01-01',
        >>>     'end_date': '2020-01-01',
        >>>     'sensors': 'L2-S2',
        >>> }
        >>> main(cmdline=cmdline, **kwargs)
    """
    import json
    # import rich
    config = StacSearchBuilderConfig.cli(cmdline=cmdline, data=kwargs,
                                         strict=True)
    # rich.print(f'config = {ub.urepr(config, nl=2)}')

    search_json = build_search_json(**ub.compatible(config, build_search_json))
    text = json.dumps(search_json, indent='    ')

    if config['out_fpath'] is not None:
        out_fpath = ub.Path(config['out_fpath'])
        out_fpath.write_text(text)
    else:
        print(text)


if __name__ == '__main__':
    """
    CommandLine:
        python ~/code/kwgis/kwgis/stac/stac_search_builder.py
    """
    main()



from concurrent.futures import as_completed
import sys
import traceback
from typing import cast

import os
import functools
import ubelt as ub
import pystac
from inspect import signature


def parallel_map_items(catalog,
                       mapper_func,
                       max_workers=4,
                       mode='process',
                       drop_on_error=True,
                       extra_args=[],
                       extra_kwargs={}):
    """
    Functions similarly to `pystac.Catalog.map_items` function but
    in parallel, and allows the mapper function to return None
    indicating that the mapped STAC item should be dropped from the
    output catalog

    Args:
        catalog (pystac.Catalog): catalog to apply transform to

        mapper_func (callable): function to be applied to each STAC item.
            The first positional argument must accept a :class:`pystac.Item`,
            and may take any arbitrary additional positional or keyword
            arguments.

        max_workers (int): number of jobs

        mode (str): process, thread, or serial

        extra_args (List | Tuple):
            extra positional args passed to mapper_func

        extra_kwargs (Dict[str, object]):
            extra keyword args passed to mapper_func

    Returns:
        pystac.Catalog: modified catalog

    Example:
        >>> # xdoctest: +SKIP("too many https errors")
        >>> from kwgis.stac.util_stac import *  # NOQA
        >>> from kwgis.demo import stac_demo
        >>> catalog_fpath = stac_demo.demo()
        >>> catalog = pystac.Catalog.from_file(catalog_fpath)
        >>> def demo_mapper_func(item):
        >>>     import copy
        >>>     print('Process: item = {}'.format(ub.urepr(item, nl=1)))
        >>>     print('item.assets = {}'.format(ub.urepr(item.assets, nl=1)))
        >>>     if 'data' not in item.assets:
        >>>         print('Drop asset without data')
        >>>         return None  # drop assets without data
        >>>     # Pretend we do some image operation and write to a new path
        >>>     in_fpath = item.assets['data'].href
        >>>     out_fpath = ub.augpath(in_fpath, suffix='_demo_process')
        >>>     print('in_fpath = {!r}'.format(in_fpath))
        >>>     out_fpath = in_fpath
        >>>     new_item = copy.deepcopy(item)
        >>>     new_item.assets['data'].href = out_fpath
        >>>     return item
        >>> out_catalog = parallel_map_items(
        >>>     catalog, demo_mapper_func, mode='serial')
        >>> assert len(list(catalog.get_all_items())) == 2, 'two items in'
        >>> assert len(list(out_catalog.get_all_items())) == 1, 'one item out'
    """
    out_catalog = catalog.full_copy()

    input_stac_items = []
    for item_link in catalog.get_item_links():
        item_link.resolve_stac_object(root=catalog.get_root())
        item = cast(pystac.Item, item_link.target)
        input_stac_items.append(item)

    executor = ub.Executor(mode=mode, max_workers=max_workers)

    jobs = [executor.submit(mapper_func, item, *extra_args, **extra_kwargs)
            for item in input_stac_items]

    output_item_links = []
    for job in as_completed(jobs):
        try:
            mapped = job.result()
        except Exception as e:
            if drop_on_error:
                print("Exception occurred (printed below), dropping item!")
                traceback.print_exception(*sys.exc_info())
                continue
            else:
                raise e
        else:
            # Allowing mappers to act as filters as well by returning None
            if mapped is None:
                continue
            elif isinstance(mapped, pystac.Item):
                output_item_links.append(pystac.Link(
                    'item', mapped, pystac.MediaType.JSON))
            else:
                for item in mapped:
                    output_item_links.append(pystac.Link(
                        'item', item, pystac.MediaType.JSON))

    out_catalog.clear_items()
    out_catalog.add_links(output_item_links)

    return out_catalog


def maps(_item_map=None, history_entry=None):
    '''
    General-purpose wrapper for STAC _item_maps.

    An _item_map should take in a STAC item and return a STAC item, an
    iterable of STAC items or None.  To support this, it should have
    an arg 'stac_item' and an arg or kwarg 'outdir'.

    This decorator handles the following tasks:
        - add original item link
        - create an item_outdir from a base outdir, and pass it to _item_map.
        - update item's self_href
        - add an entry to 'watch:process_history'

    References:
        https://pybit.es/articles/decorator-optional-argument/
        https://docs.python.org/3/library/inspect.html#inspect.BoundArguments
    '''
    if _item_map is None:
        return functools.partial(maps, history_entry=history_entry)

    if history_entry is None:
        history_entry = _item_map.__name__

    @functools.wraps(_item_map)
    def wrapper(*args, **kwargs):

        try:
            bound_args = signature(_item_map).bind(*args, **kwargs)
            bound_args.apply_defaults()
            stac_item = bound_args.arguments['stac_item']
            outdir = bound_args.arguments['outdir']
        except (IndexError, KeyError):
            raise ValueError(
                f'_item_map {_item_map.__name__} must have arg "stac_item", '
                'arg or kwarg "outdir"')

        # This assumes we're not changing the stac_item ID in any of
        # the mapping functions
        item_outdir = os.path.join(outdir, stac_item.id)
        os.makedirs(item_outdir, exist_ok=True)
        bound_args.arguments['outdir'] = item_outdir

        # Adding a reference back to the original STAC
        # item if not already present
        if len(stac_item.get_links('original')) == 0:
            stac_item.links.append(
                pystac.Link.from_dict({
                    'rel': 'original',
                    'href': stac_item.get_self_href(),
                    'type': 'application/json'
                }))
        bound_args.arguments['stac_item'] = stac_item

        output_stac_items = _item_map(*bound_args.args, **bound_args.kwargs)

        if output_stac_items is None:
            return []

        if isinstance(output_stac_items, pystac.Item):
            output_stac_items = [output_stac_items]

        for output_stac_item in output_stac_items:
            if output_stac_item is not None:
                output_stac_item.set_self_href(
                    os.path.join(item_outdir,
                                 "{}.json".format(output_stac_item.id)))

                # Roughly keeping track of what GeoWATCH processes have been
                # run on this particular item
                output_stac_item.properties.setdefault(
                    'watch:process_history', []).append(history_entry)

        return output_stac_items

    return wrapper


def associate_msi_pan(stac_catalog):
    '''
    Match up WorldView multispectral and panchromatic items.

    Returns a dict {msi_item.id: pan_item}, where pan_item can be
    nonunique.
    '''

    # Support proper stac catalog and otherwise assume the input here
    # is some iterable of STAC items
    if isinstance(stac_catalog, pystac.Catalog):
        item_list = stac_catalog.get_all_items()
    else:
        item_list = stac_catalog

    # more efficient way to do this if collections are preserved during
    # intermediate steps:
    # search = catalog.search(collections=['worldview-nitf'])
    # items = list(search.get_items()
    items_dct = {
        i.id: i
        for i in item_list
        if i.properties.get('constellation') == 'worldview'
    }

    if 0:

        # more robust way of matching PAN->MSI items one-to-many

        import pandas as pd
        from parse import parse

        df = pd.DataFrame.from_records(
            [item.properties for item in items_dct.values()])
        df['id'] = list(items_dct.keys())

        def _part(_id):
            result = parse('{}_P{part:3d}', _id)
            if result is not None:
                return result['part']
            else:
                return -1

        df['part'] = list(map(_part, df['id']))

        def _vnir_source(source):
            if source in {
                    'DigitalGlobe Acquired Image',
                    'DigitalGlobe Acquired Imagery'
            }:
                return -1
            for s in source.split(', '):
                try:
                    p = parse('{instr:l}: {rest}', s)
                    assert p['instr'] in {'SWIR', 'VNIR', 'CAVIS'}, source
                    if p['instr'] == 'VNIR':
                        return p['rest']
                except KeyError:
                    print(s, p)
            return -1

        df['vnir_source'] = list(map(_vnir_source, df['nitf:source']))

        df['geometry'] = [items_dct[i].geometry for i in df['id']]

        raise NotImplementedError

    else:

        # hacky way of matching up items by ID. Only works for pairs of 1 PAN
        # and 1 MSI, and only some sensors.
        # this matches up 40152/52563 items in the catalog.
        #
        # There are 29000 PAN items, so it matches about 2/3 of PAN.
        # The rest, besides different naming schemes, may be accounted for
        # by the fact that PAN and MSI taken during the same collect
        # can have different spatial tiling.
        mp_dct = {}
        for _id in items_dct:
            code = _id[14]
            if code != 'P':
                pid = _id[:14] + 'P' + _id[15:]
                if pid in items_dct:
                    mp_dct[_id] = items_dct[pid]
        return mp_dct



"""
Functional utilities for generating basic components of the demodata.
"""
import ubelt as ub
import kwimage
import kwarray
import numpy as np
import geopandas as gpd
from shapely import geometry
import datetime as datetime_mod
from datetime import datetime as datetime_cls


def random_geo_points(num, rng=None):
    """
    Generate a uniformly random longitude, latitude.

    Based on logic described in [SO68298220]_.

    Args:
        num (int) : number of random points to generate
        rng : random seed or number generator

    Returns:
        ndarray:
            An Nx2 array with columns corresponding to CRS84 points
            (i.e. longitude, latitude)

    References:
        .. [SO68298220] https://stackoverflow.com/questions/68298220/random-geocoords

    Example:
        >>> from kwgis.demo.metrics_demo.demo_utils import *  # NOQA
        >>> latlon = random_geo_points(num=3, rng=0)
        >>> print(ub.urepr(latlon, precision=4))
        np.array([[ 16.1579,   5.6025],
                  [-27.4843,  25.4916],
                  [ 52.5219,  11.8603]], dtype=np.float64)

    Example:
        >>> # This example demonstrates that the points are randomly spread out
        >>> from kwgis.demo.metrics_demo.demo_utils import *  # NOQA
        >>> # xdoctest: +REQUIRES(--show)
        >>> # xdoctest: +REQUIRES(module:kwplot)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> # Create random geopoints and place them into a GeoDataFrame
        >>> latlons = random_geo_points(300)
        >>> crs84 = get_crs84()
        >>> pts_gdf = gpd.GeoDataFrame(geometry=[geometry.Point(p) for p in latlons], crs=crs84)
        >>> # Plot the map of the world in the background
        >>> wld_map_gdf = gpd.read_file(
        >>>     gpd.datasets.get_path('naturalearth_lowres')
        >>> ).to_crs(crs84)
        >>> ax = wld_map_gdf.plot()
        >>> pts_gdf.plot(ax=ax, color='orange', alpha=0.8)
        >>> kwplot.show_if_requested()
    """
    rng = kwarray.ensure_rng(rng)
    u0, u1 = rng.rand(2, num)
    rad_lat = np.arcsin(2 * u0 - 1.0)  # angle with Equator   - from +pi/2 to -pi/2
    rad_lon = (2 * u1 - 1) * np.pi     # longitude in radians - from -pi to +pi
    rad_lonlat = np.stack([rad_lon, rad_lat], axis=1)
    lonlat = np.rad2deg(rad_lonlat)
    return lonlat


def random_geo_polygon(max_rt_area=10_000, rng=None):
    """
    Creates a random polygon of a "reasonable size" for a region in CRS84

    Args:
        max_rt_area (float):
            Maximum root area (i.e. sqrt(area)) of the polygon in meters.
            The generated polygon will usually have a root-area that is between
            a factor of 0.1 and 0.6 of this number.  Defaults to 10,000 meters.

        rng : random state or seed

    Returns:
        kwimage.Polygon : polygon in CRS84 space

    Example:
        >>> from kwgis.demo.metrics_demo.demo_utils import *  # NOQA
        >>> max_rt_area = 10000
        >>> region_poly = random_geo_polygon(max_rt_area, rng=321)
        >>> geo_poly = region_poly.round(6).to_geojson()
        >>> print('geo_poly = {}'.format(ub.urepr(geo_poly, nl=-1)))
        geo_poly = {
            'type': 'Polygon',
            'coordinates': [
                [
                    [-151.983974, 50.530122],
                    [-151.982547, 50.520243],
                    [-151.951446, 50.506376],
                    [-151.925555, 50.514069],
                    [-151.930728, 50.543657],
                    [-151.941043, 50.541291],
                    [-151.983974, 50.530122]
                ]
            ]
        }

    Example:
        >>> # This example demonstrates the distribution of random polygons
        >>> from kwgis.demo.metrics_demo.demo_utils import *  # NOQA
        >>> # xdoctest: +REQUIRES(--show)
        >>> # xdoctest: +REQUIRES(module:kwplot)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> # Create random polygons and place them into a GeoDataFrame
        >>> rng = kwarray.ensure_rng(32321)
        >>> # Make the polygons very large, so they show up at world scale
        >>> max_rt_area = 1_000_000
        >>> polys = [random_geo_polygon(max_rt_area, rng=rng) for _ in range(10)]
        >>> crs84 = get_crs84()
        >>> poly_gdf = gpd.GeoDataFrame(geometry=[p.to_shapely() for p in polys], crs=crs84)
        >>> # Plot the map of the world in the background
        >>> wld_map_gdf = gpd.read_file(
        >>>     gpd.datasets.get_path('naturalearth_lowres')
        >>> ).to_crs(crs84)
        >>> ax = wld_map_gdf.plot()
        >>> poly_gdf.plot(ax=ax, color='limegreen', alpha=0.8)
        >>> poly_gdf.centroid.plot(ax=ax, color='orangered', alpha=0.5)
        >>> kwplot.show_if_requested()
    """
    # Choose a random centroid of the polygon
    lon, lat = random_geo_points(num=1, rng=rng).ravel()

    # Determine what UTM zone this random centroid falls in
    epsg_code = utm_epsg_from_latlon(lat, lon)

    crs84 = get_crs84()
    pt_crs84_gdf = gpd.GeoDataFrame(geometry=[
        geometry.Point(lon, lat)], crs=crs84)

    # Project the random lon/lat point into a UTM space
    pt_utm_gdf = pt_crs84_gdf.to_crs(epsg_code)
    utm_xy = pt_utm_gdf['geometry'].iloc[0].xy
    utm_x = utm_xy[0][0]
    utm_y = utm_xy[1][0]

    # Create a random polygon with an area between 0 and 1 (usually in 0.1-0.6)
    poly = kwimage.Polygon.random(rng=rng)
    # centered it at (0, 0)
    poly = poly.translate(-np.array(poly.centroid))

    # Scale the polygon to the appropriate area in UTM space, and translate it
    # into the randomly chosen center position.
    poly_utm = poly.scale(max_rt_area).translate((utm_x, utm_y))
    utm_gdf = gpd.GeoDataFrame(geometry=[poly_utm.to_shapely()],
                               crs=pt_utm_gdf.crs)
    crs84_gdf = utm_gdf.to_crs(crs84)
    poly_crs84 = crs84_gdf['geometry'].iloc[0]
    region_poly = kwimage.Polygon.from_shapely(poly_crs84)
    return region_poly


def random_time_sequence(min_date_iso, max_date_iso, num_observations, rng=None):
    """
    Generate a list of random timestamps between the specified dates

    Args:
        min_date_iso (str): minimum possible date
        max_date_iso (str): maximum possible date
        num_observations (int): number of dates to generate
        rng: random state or seed

    Returns:
        List[datetime_cls]: sampled dates in the UTC timezone

    Example:
        >>> from kwgis.demo.metrics_demo.demo_utils import *  # NOQA
        >>> min_date_iso = '1996-06-23'
        >>> max_date_iso = '2017-10-27'
        >>> num_observations = 3
        >>> obs_sequence = random_time_sequence(min_date_iso, max_date_iso, num_observations, rng=9320)
        >>> print('obs_sequence = {}'.format(ub.urepr(obs_sequence, nl=1)))
        obs_sequence = [
            datetime.datetime(1997, 9, 17, 20, 49, 7, 888653, tzinfo=datetime.timezone.utc),
            datetime.datetime(2001, 6, 29, 17, 43, 36, 108233, tzinfo=datetime.timezone.utc),
            datetime.datetime(2009, 4, 2, 10, 26, 54, 429200, tzinfo=datetime.timezone.utc),
        ]

    Ignore:
        # In Docker, test that the code provides the same number in different
        # locale.
        export TZ="Africa/Lusaka"
        rm -rf /etc/localtime
        ln -s /usr/share/zoneinfo/Africa/Lusaka /etc/localtime
        python -c "import datetime as datetime_mod; print(datetime_mod.datetime.fromisoformat('1996-06-23').replace(tzinfo=datetime_mod.timezone.utc).timestamp())"
    """
    rng = kwarray.ensure_rng(rng)
    from kwutil import util_time
    min_date = util_time.coerce_datetime(min_date_iso)
    max_date = util_time.coerce_datetime(max_date_iso)
    # min_date = datetime_cls.fromisoformat(min_date_iso)
    # max_date = datetime_cls.fromisoformat(max_date_iso)
    # # Assume user is specifying UTC by default.
    # if min_date.tzinfo is None:
    #     min_date = min_date.replace(tzinfo=datetime_mod.timezone.utc)
    # if max_date.tzinfo is None:
    #     max_date = max_date.replace(tzinfo=datetime_mod.timezone.utc)
    min_ts = min_date.timestamp()
    max_ts = max_date.timestamp()
    obs_timestamps = rng.rand(num_observations) * (max_ts - min_ts) + min_ts
    obs_sequence = [
        # util_time.coerce_datetime(ts)
        datetime_cls.fromtimestamp(ts, tz=datetime_mod.timezone.utc)
        for ts in sorted(obs_timestamps)
    ]
    return obs_sequence


def utm_epsg_from_latlon(lat, lon):
    """
    Find a reasonable UTM CRS for a given lat / lon

    The purpose of this function is to get a reasonable CRS for computing
    distances in meters. If the region of interest is very large, this may not
    be valid.

    See [SE190198]_ and [SE365584]_.

    Args:
        lat (float): degrees in latitude
        lon (float): degrees in longitude

    Returns:
        int : the ESPG code of the UTM zone

    References:
        .. [SE190198] https://gis.stackexchange.com/questions/190198/how-to-get-appropriate-crs-for-a-position-specified-in-lat-lon-coordinates
        .. [SE365584] https://gis.stackexchange.com/questions/365584/convert-utm-zone-into-epsg-code

    Example:
        >>> from kwgis.demo.metrics_demo.demo_utils import *  # NOQA
        >>> epsg_code = utm_epsg_from_latlon(0, 0)
        >>> print('epsg_code = {!r}'.format(epsg_code))
        epsg_code = 32631
    """
    import utm
    # easting, northing, zone_num, zone_code = utm.from_latlon(min_lat, min_lon)
    zone_num = utm.latlon_to_zone_number(lat, lon)

    # Construction of EPSG code from UTM zone number
    south = lat < 0
    epsg_code = 32600
    epsg_code += int(zone_num)
    if south is True:
        epsg_code += 100
    return epsg_code


def project_gdf_to_local_utm(gdf_crs84, max_utm_zones=None):
    """
    Find the local UTM zone for a geo data frame and project to it.

    Assumes geometry is in CRS-84.

    All geometry in the GDF must be in the same UTM zone.

    Args:
        gdf_crs84 (geopandas.GeoDataFrame):
            The data with CRS-84 geometry to project into a local UTM

        max_utm_zones (int | None):
            If the data spans more than this many UTM zones, error.
            Otherwise, we take the first one.

    Returns:
        geopandas.GeoDataFrame

    Example:
        >>> import geopandas as gpd
        >>> import kwarray
        >>> import kwimage
        >>> rng = kwarray.ensure_rng(0)
        >>> # Gen lat/lons between 0 and 1, which is in UTM zone 31N
        >>> gdf_crs84 = gpd.GeoDataFrame({'geometry': [
        >>>     kwimage.Polygon.random(rng=rng).to_shapely(),
        >>>     kwimage.Polygon.random(rng=rng).to_shapely(),
        >>>     kwimage.Polygon.random(rng=rng).to_shapely(),
        >>> ]}, crs=get_crs84())
        >>> gdf_utm = project_gdf_to_local_utm(gdf_crs84)
        >>> assert gdf_utm.crs.name == 'WGS 84 / UTM zone 31N'

    Example:
        >>> import geopandas as gpd
        >>> import kwarray
        >>> import kwimage
        >>> # If the data is too big for a single UTM zone,
        >>> rng = kwarray.ensure_rng(0)
        >>> gdf_crs84 = gpd.GeoDataFrame({'geometry': [
        >>>     kwimage.Polygon.random(rng=rng).scale(90).to_shapely(),
        >>>     kwimage.Polygon.random(rng=rng).scale(90).to_shapely(),
        >>>     kwimage.Polygon.random(rng=rng).scale(90).to_shapely(),
        >>> ]}, crs=get_crs84())
        >>> import pytest
        >>> with pytest.raises(ValueError):
        >>>     gdf_utm = project_gdf_to_local_utm(gdf_crs84, max_utm_zones=1)
    """
    epsg_zones = []
    for geom_crs84 in gdf_crs84.geometry:
        epsg_zone = find_local_meter_epsg_crs(geom_crs84)
        epsg_zones.append(epsg_zone)
    if not ub.allsame(epsg_zones):
        unique_utm = set(epsg_zones)
        if max_utm_zones is not None and len(unique_utm) > max_utm_zones:
            raise ValueError(ub.paragraph(
                '''
                Input data spanned multiple UTM zones.
                This is currently not allowed. {}
                '''
            ).format(unique_utm))
    # TODO: if there are more than one is there a way to get "the best one?"
    epsg_zone = epsg_zones[0]
    gdf_utm = gdf_crs84.to_crs(epsg_zone)
    return gdf_utm


def find_local_meter_epsg_crs(geom_crs84):
    """
    Find the "best" meter based CRS for a smallish geographic region.

    Currently this only returns UTM zones. Might be better to return an Albers
    projection if the geometry spans more than one UTM zone.

    Args:
        geom_crs84 (shapely.geometry.base.BaseGeometry):
            shapely geometry in CRS84 (lon/lat wgs84)

    Returns:
        int: epsg code

    References:
        [1] https://gis.stackexchange.com/questions/148181/choosing-projection-crs-for-short-distance-based-analysis/148187
        [2] http://projfinder.com/

    TODO:
        - [ ] Better UTM zone intersection
        - [ ] Fix edge cases

    Example:
        >>> import kwimage
        >>> geom_crs84 = kwimage.Polygon.random().translate(-0.5).scale((180, 90)).to_shapely()
        >>> epsg_zone = find_local_meter_epsg_crs(geom_crs84)
    """
    lonmin, latmin, lonmax, latmax = geom_crs84.bounds
    # Hack: this doesnt work on boundries (or for larger regions)
    # correct way of doing this would be lookup candiate CRS zones,
    # and find the one with highest intersection area weighted by distance
    # to the center of the valid region.
    latmid = (latmin + latmax) / 2
    lonmid = (lonmin + lonmax) / 2
    candidate_utm_codes = [
        utm_epsg_from_latlon(latmin, lonmin),
        utm_epsg_from_latlon(latmax, lonmax),
        utm_epsg_from_latlon(latmax, lonmin),
        utm_epsg_from_latlon(latmin, lonmax),
        utm_epsg_from_latlon(latmid, lonmid),
    ]
    epsg_zone = ub.argmax(ub.dict_hist(candidate_utm_codes))
    return epsg_zone


@ub.memoize
def get_crs84():
    """
    Constructing the CRS84 is slow.
    This function memoizes it so it only happens once.

    Returns:
        pyproj.crs.crs.CRS
    """
    from pyproj import CRS
    crs84 = CRS.from_user_input('OGC:CRS84')
    # crs84 = CRS.from_user_input('crs84')
    return crs84



r"""
Helpers for reading data downloaded from digital globe


Notes:
    The data in the Core3D dataset is public and can be rehosted.

    https://spacenet.ai/core3d/

    AWS_PROFILE=iarpa aws s3 ls s3://spacenet-dataset/Hosted-Datasets/CORE3D-Public-Data/Satellite-Images/
    AWS_PROFILE=iarpa aws s3 ls s3://spacenet-dataset/Hosted-Datasets/CORE3D-Public-Data/Tiled-Examples-for-Urban-3D-Challenge-Comparisons/02_Master/challenge-inputs/

    AWS_PROFILE=iarpa aws s3 cp \
            s3://spacenet-dataset/Hosted-Datasets/CORE3D-Public-Data/Tiled-Examples-for-Urban-3D-Challenge-Comparisons/02_Master/challenge-inputs/RIC_Tile_000_DSM.tif RIC_Tile_000_DSM.tif
    AWS_PROFILE=iarpa aws s3 cp \
            s3://spacenet-dataset/Hosted-Datasets/CORE3D-Public-Data/Tiled-Examples-for-Urban-3D-Challenge-Comparisons/02_Master/challenge-inputs/RIC_Tile_000_DTM.tif RIC_Tile_000_DTM.tif
    AWS_PROFILE=iarpa aws s3 cp \
            s3://spacenet-dataset/Hosted-Datasets/CORE3D-Public-Data/Tiled-Examples-for-Urban-3D-Challenge-Comparisons/02_Master/challenge-inputs/RIC_Tile_000_RGB.tif RIC_Tile_000_RGB.tif

Requirements:
    pip install xmltodict
    pip install pyshp
    pip install cogeotiff
"""
from os.path import exists
from os.path import join
import ubelt as ub
import xmltodict
from os.path import dirname, abspath


class DigitalGlobeBundle(ub.NiceRepr):
    """
    Data structure to organize information in digital globe bundles

    TODO: need public digital globe demodata for a doctest

    Maybe we can grab them from here?
    https://www.maxar.com/product-samples
    https://ard.maxar.com/samples#v5/

    https://spacenet.ai/core3d/

    Requirements:
        pip isntall pyshp

    Ignore:
        # This has a different format than our stuff... bleh..
        sample_zip_fpath = ub.grabdata('https://maxar-marketing.s3.amazonaws.com/product-samples/Rome_Colosseum_2022-03-22_WV03_HD.zip', hash_prefix='2a99cea2b37bed9b5867fa21a1bd')
        from kwcoco.util import util_archive
        archive = util_archive.Archive(sample_zip_fpath)
        dpath = (ub.Path(sample_zip_fpath).parent / 'MaxarSample').ensuredir()
        metadata_fpath = list(dpath.glob('*.MAN'))[0]
        archive.extractall(dpath)
        delivery_metadata_fpath = dpath / '050012575010_01/050012575010_01_README.XML'
        self = DigitalGlobeBundle(delivery_metadata_fpath)
    """

    def __init__(self, delivery_metadata_fpath, pointer=None, autobuild=True):
        self.data = {
            'delivery_metadata_fpath': delivery_metadata_fpath,
            'product_metas': None,
            'pointer': pointer,
        }
        if autobuild:
            self.parse_delivery_metadata()

    def __nice__(self):
        return self.data['delivery_metadata_fpath']

    def parse_delivery_metadata(self):
        import shapefile
        import kwimage
        delivery_metadata_fpath = self.data['delivery_metadata_fpath']
        dpath = dirname(delivery_metadata_fpath)

        with open(delivery_metadata_fpath, 'r') as file:
            delivery_metadata = xmltodict.parse(file.read())

        self.data['other'] = ub.dict_diff(delivery_metadata, {'DeliveryMetadata'})
        self.data['non_product'] = ub.dict_diff(delivery_metadata, {'DeliveryMetadata'})

        product_list = delivery_metadata['DeliveryMetadata']['product']

        pointer = self.data['pointer']

        product_metas = []
        for product in product_list:
            product_meta = product.copy()
            prod_files = product_meta.pop('productFile')

            # Find the files associated with the order AOI
            aoi_fpaths = {
                'shp': None,
                'dbf': None,
                'shx': None,
                'prj': None,
            }
            misc_exts = {
                '_LAYOUT.JPG',
                'NEXTVIEW.TXT',
                '_README.TXT',
                '_README.XML',
                '-BROWSE.JPG',
            }
            prod_types = []
            for v in prod_files:
                product_type = None

                for ext in misc_exts:
                    if v['filename'].lower().endswith(ext.lower()):
                        product_type = 'misc'

                if product_type is None:
                    for ext in aoi_fpaths.keys():
                        fname = 'ORDER_SHAPE.' + ext
                        if v['filename'].lower().endswith(fname.lower()):
                            fpath = join(dpath, v['relativeDirectory'], v['filename'])
                            assert aoi_fpaths[ext] is None
                            aoi_fpaths[ext] = fpath
                            product_type = 'aoi'

                if product_type is None:
                    if v['relativeDirectory'].lower().endswith('GIS_FILES'.lower()):
                        product_type = 'misc-gis'

                if product_type is None:
                    for ext in kwimage.im_io.IMAGE_EXTENSIONS:
                        if v['filename'].lower().endswith(ext.lower()):
                            product_type = 'image'

                if product_type is None:
                    product_type = 'other'

                prod_types.append(product_type)

            type_to_prods = ub.group_items(prod_files, prod_types)
            type_to_prods['image']
            type_to_prods['misc']
            type_to_prods['gis']
            type_to_prods['aoi']
            type_to_prods['other']
            # prod_type_hist = ub.map_vals(len, type_to_prods)
            # print('prod_type_hist = {}'.format(ub.urepr(prod_type_hist, nl=1)))

            aoi_files = {key: open(val, 'rb') for key, val in aoi_fpaths.items()}
            try:
                shp_wkt = ensure_unicode(aoi_files['prj'].read())
                shp_reader = shapefile.Reader(
                    shp=aoi_files['shp'],
                    dbf=aoi_files['dbf'],
                    shx=aoi_files['shx']
                )
                aoi_geojson = shp_reader.shape().__geo_interface__
                product_meta['aoi_geojson'] = aoi_geojson
                product_meta['shp_wkt'] = shp_wkt
            finally:
                for val in aoi_files.values():
                    val.close()
                shp_reader.close()

            if pointer is not None:
                for v in type_to_prods['image']:
                    prod_fname = v['filename']
                    flag = pointer.endswith(prod_fname)
                    v['is_pointer'] = flag

            product_meta['images'] = type_to_prods['image']
            product_metas.append(product_meta)

        self.data['product_metas'] = product_metas

    @classmethod
    def from_pointer(cls, pointer, **kw):
        """
        Args:
            pointer (str): a path to any file inside a digital globe
                bundle. We will search for the DeliveryMetadata.xml data.

        Ignore:
            pointer = '/home/joncrall/data/dvc-repos/smart_watch_dvc/drop0/KR-Pyeongchang-WV/_assets/20170907_a_KRP_011777481_10_0/011777481010_01_003/011777481010_01/011777481010_01_P001_MUL/17SEP07021826-M1BS-011777481010_01_P001.TIF'

            pointer = '/home/joncrall/data/dvc-repos/smart_watch_dvc/drop0/KR-Pyeongchang-WV/_assets/20170907_a_KRP_011777481_10_0/011777481010_01_003/011777481010_01/011777481010_01_P001_PAN/17SEP07021826-P1BS-011777481010_01_P001.TIF'

            cls = DigitalGlobeBundle
            self = DigitalGlobeBundle.from_pointer(pointer)

            for meta in self.data['product_metas']:
                meta['sensorVehicle']

            dict_list = self.data['product_metas']
            print(varried['sensorVehicle'])
        """
        dpath = abspath(pointer)
        delivery_fpath = search_path_ancestors(
            path=dpath, fname='DeliveryMetadata.xml')
        if delivery_fpath is None:
            raise Exception('cannot find DG DeliveryMetadata.xml')

        self = cls(delivery_fpath, pointer=pointer, **kw)
        return self

    @classmethod
    def coerce(cls, key, **kw):
        try:
            self = cls.pointer(key, **kw)
        except Exception:
            self = None
        return self


def search_path_ancestors(path, fname, stop_fname=None, max_steps=1000):
    """
    Search path and all of its containing folders for a file name ``fname``.

    Args:
        path (str): directory to start the search
        fname (str): path to search for
        stop_fname (str): stop if we find a file with this name.
    """
    import itertools as it
    dpath = path
    found = None
    for idx in it.count():
        fpath = join(dpath, fname)
        if exists(fpath):
            found = fpath
            break
        if stop_fname is not None:
            stop_fpath = join(dpath, stop_fname)
            if exists(stop_fpath):
                raise Exception('found stop fname, cannot find {}'.format(fname))
        dpath_next = dirname(dpath)
        if idx > max_steps:
            raise Exception('too many steps, cannot find {}'.format(fname))
        if dpath_next == dpath:
            raise Exception('reached the root, cannot find {}'.format(fname))
        dpath = dpath_next
    return found


def ensure_unicode(text):
    r"""
    Casts bytes into utf8 (mostly for python2 compatibility).

    Args:
        text (str | bytes):
            text to ensure is decoded as unicode

    Returns:
        str

    References:
        .. [SO_12561063] http://stackoverflow.com/questions/12561063/extract-data-from-file

    Example:
        >>> import codecs  # NOQA
        >>> assert ensure_unicode('my ünicôdé strįng') == 'my ünicôdé strįng'
        >>> assert ensure_unicode('text1') == 'text1'
        >>> assert ensure_unicode('text1'.encode('utf8')) == 'text1'
        >>> assert ensure_unicode('ï»¿text1'.encode('utf8')) == 'ï»¿text1'
        >>> assert (codecs.BOM_UTF8 + 'text»¿'.encode('utf8')).decode('utf8')
    """
    if isinstance(text, str):
        return text
    elif isinstance(text, bytes):
        return text.decode('utf8')
    else:  # nocover
        raise ValueError('unknown input type {!r}'.format(text))



class MetadataNotFound(Exception):
    """
    Thrown when metadata does not exist
    """
    pass


class GeoMetadataNotFound(MetadataNotFound):
    """
    Thrown when geographic metadata does not exist
    """
    pass
